#include "OpenGL.h"

#pragma comment( lib, "opengl32.lib" )							// Search For OpenGL32.lib While Linking
#pragma comment( lib, "glu32.lib" )								// Search For GLu32.lib While Linking
#pragma comment( lib, "glaux.lib" )								// Search For GLaux.lib While Linking

void GLSetProjection(const Camera& cam)
{
	double d[16];
	const Matrixf3& m = cam.transform_matrix;
	double m_a_z = m.a.z + cam.z_delta;
	d[ 0]  = m.x.x / cam.view_px; d[ 1] = -m.x.y / cam.view_py; d[ 2] = m.x.z; d[ 3] = m.x.z;
	d[ 4]  = m.y.x / cam.view_px; d[ 5] = -m.y.y / cam.view_py; d[ 6] = m.y.z; d[ 7] = m.y.z;
	d[ 8]  = m.z.x / cam.view_px; d[ 9] = -m.z.y / cam.view_py; d[10] = m.z.z; d[11] = m.z.z;
	d[ 12] = m.a.x / cam.view_px; d[13] = -m.a.y / cam.view_py; d[14] = m_a_z; d[15] = m.a.z;

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixd(d);
}

PaintGL::PaintGL()
{
	size = Null;
#ifdef PLATFORM_WIN32
	hbmp = ohbmp = NULL;
	hdc = NULL;
	hrc = NULL;
#endif
#ifdef PLATFORM_X11
	glxcxt = NULL;
	glxbmp = 0;
	pixmap = 0;
#endif
}

PaintGL::~PaintGL()
{
	Free();
}

void PaintGL::Free()
{
#ifdef PLATFORM_WIN32
	if(hrc) {
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(hrc);
		SelectObject(hdc, ohbmp);
		DeleteDC(hdc);
		DeleteObject(hbmp);
		hrc = NULL;
		hbmp = ohbmp = NULL;
		hdc = NULL;
		hbmp = NULL;
	}
#endif
#ifdef PLATFORM_X11
	if(glxbmp) {
		glXDestroyGLXPixmap(Xdisplay, glxbmp);
		glxbmp = 0;
	}
	if(pixmap) {
		XFreePixmap(Xdisplay, pixmap);
		pixmap = 0;
	}
	if(glxcxt) {
		glXDestroyContext(Xdisplay, glxcxt);
		glxcxt = NULL;
	}
#endif
	size = Null;
}


void PaintGL::Init(Size sz) {
	Free();

	size = sz;

#ifdef PLATFORM_WIN32
	BITMAPINFOHEADER bih;
	memset(&bih, 0, sizeof(bih));
	bih.biSize = sizeof(bih);
	bih.biWidth = ((((int) sz.cx * 8) + 31) & ~31) >> 3;
	bih.biHeight = sz.cy;
	bih.biPlanes = 1;
	bih.biBitCount = 32;
	bih.biCompression = BI_RGB;

	hdc = CreateCompatibleDC(NULL);

	void *dummy;

	hbmp = CreateDIBSection(hdc, (BITMAPINFO*)&bih, DIB_PAL_COLORS, &dummy, NULL, 0);

	ohbmp = (HBITMAP)SelectObject(hdc, hbmp);

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(pfd));
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL/* | PFD_SUPPORT_GDI | PFD_ACCELERATED*/;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 32;
	pfd.iLayerType = PFD_MAIN_PLANE;

	GLuint PixelFormat = ChoosePixelFormat(hdc, &pfd);
	SetPixelFormat(hdc, PixelFormat, &pfd);
	hrc = wglCreateContext(hdc);
	wglMakeCurrent(hdc, hrc);
#endif

#ifdef PLATFORM_X11
	pixmap = XCreatePixmap(Xdisplay, Xroot, sz.cx, sz.cy, Xdepth);
	ASSERT(pixmap);
	int Xattriblist[] = {
		GLX_DEPTH_SIZE, 8,
		GLX_RGBA,
		GLX_RED_SIZE, 4,
		GLX_GREEN_SIZE, 4,
		GLX_BLUE_SIZE, 4,
		None
	};
	XVisualInfo *vinfo = glXChooseVisual(Xdisplay, Xscreenno, Xattriblist);
	ASSERT(vinfo);
	glxbmp = glXCreateGLXPixmap(Xdisplay, vinfo, pixmap);
	ASSERT(glxbmp);
	glxcxt = glXCreateContext(Xdisplay, vinfo, NULL, false);
	XFree(vinfo);
	glXMakeCurrent(Xdisplay, glxbmp, glxcxt);
#endif

	glShadeModel(GL_SMOOTH);
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glViewport(0, 0, (GLsizei)sz.cx, (GLsizei)sz.cy);
//	glMatrixMode(GL_PROJECTION);
//	glLoadIdentity();
//	gluPerspective(45.0f, (GLfloat)(sz.cx)/(GLfloat)(sz.cy), 1.0f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void PaintGL::Paint(Draw& w, const Rect& r, Callback gl)
{
	if(r.Size() != size)
		Init(r.Size());
	gl();
	glFlush();
#ifdef PLATFORM_WIN32
	HDC whdc = w.BeginGdi();
	BitBlt(whdc, 0, 0, size.cx, size.cy, hdc, r.left, r.top, SRCCOPY);
	w.EndGdi();
#endif
#ifdef  PLATFORM_X11
	XCopyArea(Xdisplay, pixmap, w.GetDrawable(), w.GetGC(), 0, 0, size.cx, size.cy, r.left, r.top);
#endif
}

CameraCtrl::CameraCtrl()
{
	Color ang_bg(192, 255, 255), len_bg(255, 255, 192);
//	zenith.MinMax(Null, Null);
	zenith.SetLabel("Zenith ()");
	zenith.Vert(false);
	zenith.AutoSelect();
	zenith.SetBackground(ang_bg);
	elevation.SetLabel("Elevation (m)");
	elevation.Vert(true);
	elevation.AutoSelect();
	elevation.SetBackground(len_bg);
	distance.SetLabel("Distance (m)");
	distance.Horz(false);
	distance.AutoSelect();
	distance.SetBackground(len_bg);
//	elevation.MinMax(-1000, +1000);
	azimuth.SetLabel("Azimuth ()");
	azimuth.Horz(true);
	azimuth.AutoSelect();
	azimuth.SetBackground(ang_bg);
//	azimuth.MinMax(-180, +180);
	AddFrame(zenith);
	AddFrame(elevation);
	AddFrame(distance);
	AddFrame(azimuth);
	elevation <<= 0;
	azimuth <<= 0;
	zenith <<= 30;
	distance <<= 200;
	elevation <<= azimuth <<= zenith <<= distance <<= THISBACK1(CalculateView, false);
	target = Pointf3(0, 0, 0);
	MakeHill();
}

static void SetStep(RulerCtrl& ruler, const int *choices, int nchoices, int text_gap = 40)
{
	Size sz = ruler.GetSize();
	int len = (ruler.IsVert() ? sz.cy : sz.cx);
	double sc = fabs(ruler.GetScale());
	while(--nchoices > 0 && choices[nchoices - 1] * sc >= text_gap)
		;
	int t = choices[nchoices];
	ruler.SetTextStep(t);
	int i = 0;
	while(i < nchoices && choices[i] * sc < 4)
		i++;
	while(i < nchoices && t % choices[i])
		i++;
	ruler.SetSmallStep(choices[i]);
}

static void SetAngStep(RulerCtrl& ruler)
{
	static const int ang_step[] = { 1, 2, 3, 5, 10, 15, 30, 45 };
	SetStep(ruler, ang_step, __countof(ang_step), 30);
}

static void SetLenStep(RulerCtrl& ruler)
{
	static const int len_step[] = { 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000, };
	SetStep(ruler, len_step, __countof(len_step));
}

void CameraCtrl::Layout()
{
	int zy = zenith.GetSize().cy, ey = elevation.GetSize().cy, ax = azimuth.GetSize().cx, dx = distance.GetSize().cx;
	zenith.SetZoom(-(zy - 40) / 360.0, zy >> 1);
	elevation.SetZoom(-ey / 1000.0, ey >> 1);
	azimuth.SetZoom((ax - 40) / 360.0, ax >> 1);
	distance.SetZoom(dx / 1000.0, 0);
	SetAngStep(zenith);
	SetAngStep(azimuth);
	SetLenStep(elevation);
	SetLenStep(distance);
	CalculateView();
}

void CameraCtrl::Paint(Draw& draw)
{
	glpaint.Paint(draw, GetSize(), THISBACK(GLPaint));
	Rect rc(GetSize());
//	draw.DrawRect(rc, SWhite());
	draw.Offset(rc.CenterPoint());
	for(int x = 0; x < SIDE; x++)
		for(int y = 0; y < SIDE; y++) {
			Pointf3 a((x - HALF) * 10, (y - HALF) * 10, hfield[x][y]), b = a;
			if(x < SIDE - 1) {
				Pointf3 c(a.x + 10, a.y, hfield[x + 1][y]);
				Pointf t, u;
				if(camera.TransformClip(a, c, t, u)) draw.DrawLine(t, u, 0, Blue());
			}
			if(y < SIDE - 1) {
				Pointf3 d(a.x, a.y + 10, hfield[x][y + 1]);
				Pointf t, u;
				if(camera.TransformClip(b, d, t, u)) draw.DrawLine(t, u, 0, Green());
			}

		}
	draw.End();
}

void CameraCtrl::GLPaint()
{
	Pointf3 litedir = Unit(Pointf3(1, 1, 1));
	glClearColor(1, 1, 1, 10);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	GLSetProjection(camera);
	for(int x = 1; x < SIDE; x++) {
		glBegin(GL_QUAD_STRIP);
		for(int y = 0; y < SIDE; y++) {
			GLColor(Color(255, normal[x - 1][y], 0));
			GLVertex(Pointf3((x - 1 - HALF) * 10, (y - HALF) * 10, hfield[x - 1][y]));
			GLColor(Color(255, normal[x - 0][y], 0));
			GLVertex(Pointf3((x - 0 - HALF) * 10, (y - HALF) * 10, hfield[x - 0][y]));
		}
		glEnd();
	}
	glFinish();
}

Pointf3 CameraCtrl::CalcDragTarget(Point start, Point end) const
{
	if(IsNull(end))
		return target;
	Size delta = end - start;
	return target + drag_right * delta.cx + drag_down * delta.cy;
}

bool CameraCtrl::Push(Point pt, dword keyflags)
{
	drag_camera = camera;
	Pointf3 base = camera.Transform(camera.target);
	Pointf3 right = camera.Transform(camera.target + camera.straight_right) - base;
	Pointf3 up = camera.Transform(camera.target + camera.straight_up) - base;
	drag_right = camera.straight_right / -Length(right);
	drag_down = camera.straight_up / -Length(up);
	target = camera.target;
	return true;
}

void CameraCtrl::Drag(Point start, Point old_pt, Point new_pt, dword keyflags)
{
	drag_target = target;
	drag_azimuth = Nvl((double)~azimuth);
	drag_zenith = Nvl((double)~zenith);
	if(keyflags & K_MOUSELEFT)
		drag_target = CalcDragTarget(start, new_pt);
	else if((keyflags & K_MOUSERIGHT) && !IsNull(new_pt)) {
		Size delta = new_pt - start;
		drag_azimuth += delta.cx * 360.0 / GetSize().cx;
		drag_zenith += delta.cy * 360.0 / GetSize().cy;
	}
	CalculateView(true);
}

void CameraCtrl::Drop(Point start, Point end, dword keyflags)
{
	if(keyflags & K_MOUSELEFT)
		target = CalcDragTarget(start, end);
	else if(keyflags & K_MOUSERIGHT) {
		double a = Nvl((double)~azimuth);
		double z = Nvl((double)~zenith);
		Size delta = end - start;
		a += delta.cx * 360.0 / GetSize().cx;
		z += delta.cy * 360.0 / GetSize().cy;
		azimuth <<= modulo(a + 180, 360.0) - 180;
		zenith <<= modulo(z + 180, 360.0) - 180;
	}
	CalculateView(false);
}

void CameraCtrl::CalculateView(bool dragging)
{
	double z = (dragging ? drag_zenith : Nvl((double)~zenith, 30.0)) * DEGRAD;
	double a = (dragging ? drag_azimuth : Nvl((double)~azimuth, 0.0)) * DEGRAD;
	double d = minmax(Nvl((double)~distance, 1000.0), 10.0, 100000.0);
	double e = Nvl((double)~elevation, 0.0);
	Pointf3 trg = (dragging ? drag_target : target) - Pointf3(0, 0, e);
	Pointf3 loc = trg + RotateZ(RotateX(Pointf3(0, -d, 0), z), -a);
	Pointf3 up = RotateZ(RotateX(Pointf3(0, 0, 1), z), -a);
	camera.Location(loc);
	camera.Target(trg);
	camera.Upwards(up);
	Size sz = GetSize();
	Size mmsize = iscale(sz, Size(1000, 1000), GetPixelsPerMeter(ScreenInfo()));
	camera.Viewport(mmsize);
	camera.ViewportSize(sz >> 1);
	camera.NearDistance(1);
	camera.FarDistance(10000);
	camera.Update();
	Refresh();

	Pointf3 src(10, 10, 0);
	Pointf3 prj = camera.Transform(src);
	Pointf3 src2 = camera.InverseXY(Pointf(prj.x, prj.y), 0);
}

/*
Pointf3 CameraCtrl::DragInverse(Point pt) const
{
	if(IsNull(pt))
		return Null;
	pt -= GetSize() >> 1;
	Pointf3 in = drag_camera.InverseXY(Pointf(pt), drag_camera.target.z);
	if(!IsNull(in)) {
		Pointf3 out = drag_camera.Transform(in);
		DUMP(pt);
		DUMP(in);
		DUMP(out);
	}
	return in;
}
*/

void CameraCtrl::AddHill(int x, int y, double height, double radius)
{
	for(int i = 0; i < SIDE; i++)
		for(int j = 0; j < SIDE; j++) {
			double d = sqrt(sqr(i - x) + sqr(j - y));
			double h = height * (1 - min(d, radius) / radius);
			hfield[i][j] += h;
		}
}

void CameraCtrl::MakeHill()
{
	ZeroArray(hfield);
	for(int level = 0; level < 6; level++) {
		int pw4 = 1 << (2 * level);
		double sq = 1 + level;
		sq = sq * sq * sq;
		for(int i = 0; i < pw4; i++) {
			AddHill(rand() % SIDE, rand() % SIDE, (rand() % 400 - 200) / (double)pw4, SIDE / (double)pw4);
		}
	}
	CalcNormals();
}

void CameraCtrl::CalcNormals()
{
	Pointf3 litedir = Unit(Pointf3(1, 1, 1));
	for(int i = 0; i < SIDE; i++)
		for(int j = 0; j < SIDE; j++) {
			Pointf3 b(10 * i, 10 * j, hfield[i][j]), xp, xm, yp, ym;
			xp = xm = yp = ym = b;
			if(i + 1 < SIDE) xp = Pointf3(b.x + 10, b.y, hfield[i + 1][j]);
			if(i > 0)        xm = Pointf3(b.x - 10, b.y, hfield[i - 1][j]);
			if(j + 1 < SIDE) yp = Pointf3(b.x, b.y + 10, hfield[i][j + 1]);
			if(j > 0)        ym = Pointf3(b.x, b.y - 10, hfield[i][j - 1]);
			Pointf3 n = Unit((xp - xm) % (yp - ym));
			double d = litedir ^ n;
			normal[i][j] = minmax(fround(d * 96 + 160), 0, 255);
		}
}
