#ifndef _OpenGL_OpenGL_h
#define _OpenGL_OpenGL_h

#include <Geom/Ctrl/GeomCtrl.h>

#include <GL/gl.h>
#include <GL/glu.h>
#ifdef PLATFORM_WIN32
#include <GL/glaux.h>
#endif
#ifdef PLATFORM_X11
#define Display XDisplay
#include <GL/glx.h>
#undef Display
#endif

class PaintGL {
	Size    size;
#ifdef PLATFORM_WIN32
	HBITMAP hbmp, ohbmp;
	HDC     hdc;
	HGLRC   hrc;
#endif
#ifdef PLATFORM_X11
	GLXContext glxcxt;
	GLXPixmap  glxbmp;
	Pixmap     pixmap;
#endif

	void Free();
	void Init(Size sz);

public:
	void Paint(Draw& w, const Rect& r, Callback gl);

	PaintGL();
	~PaintGL();
};

inline void GLColor(Color c)           { glColor3ub(c.GetR(), c.GetG(), c.GetB()); }
inline void GLVertex(const Pointf3& v) { glVertex3dv(&v.x); }

class CameraCtrl : public DragDropCtrl {
public:
	typedef CameraCtrl CLASSNAME;
	CameraCtrl();

	virtual void Layout();
	virtual void Paint(Draw& draw);

	virtual bool Push(Point pt, dword keyflags);
	virtual void Drag(Point start, Point old_pt, Point new_pt, dword keyflags);
	virtual void Drop(Point start, Point end, dword keyflags);

	void         CalculateView(bool dragging = false);

//	Pointf3      DragInverse(Point pt) const;
	Pointf3      CalcDragTarget(Point start, Point end) const;

public:
	Camera       camera;
	RulerCtrl    zenith;
	RulerCtrl    azimuth;
	RulerCtrl    elevation;
	RulerCtrl    distance;
	Pointf3      target;
	Pointf3      drag_target;
	double       drag_azimuth;
	double       drag_zenith;

private:
	Pointf3      drag_start;
	Pointf3      drag_end;
	Camera       drag_camera;
	Pointf3      drag_right;
	Pointf3      drag_down;
	PaintGL      glpaint;

private:
	// test image
	enum { HALF = 10, SIDE = 2 * HALF };
	double       hfield[SIDE][SIDE];
	int          normal[SIDE][SIDE];

	void         AddHill(int x, int y, double height, double radius);
	void         MakeHill();
	void         CalcNormals();
	void         GLPaint();
};

#endif
