#ifdef _MT

#ifdef PLATFORM_WIN32

typedef LONG Atomic;

inline int  AtomicRead(const Atomic& t)      { return t; }
inline void AtomicWrite(Atomic& t, int val)  { t = val; }

inline int  AtomicInc(Atomic& t)             { return InterlockedIncrement(&t); }
inline int  AtomicDec(Atomic& t)             { return InterlockedDecrement(&t); }
inline int  AtomicXAdd(Atomic& t, int incr)  { return InterlockedExchangeAdd(&t, incr); }

class CriticalSection {
protected:
	CRITICAL_SECTION section;

	CriticalSection(int)         {}

	friend CriticalSection& MainCriticalSection();

public:
	void  Enter()                { EnterCriticalSection(&section); }
	void  Leave()                { LeaveCriticalSection(&section); }

	CriticalSection()            { InitializeCriticalSection(&section); }
	~CriticalSection()           { DeleteCriticalSection(&section); }

	struct Lock;
	struct LockMain;
};

#endif

#ifdef PLATFORM_POSIX

#include <bits/atomicity.h>

typedef _Atomic_word Atomic;

// Really just on x86...

inline int  AtomicRead(const Atomic& t)      { return t; } 
inline void AtomicWrite(Atomic& t, int val)  { t = val; }

inline int  AtomicInc(Atomic& t)             { __atomic_add(&t, 1); return t; }
inline int  AtomicDec(Atomic& t)             { __atomic_add(&t, -1); return t; }
inline int  AtomicXAdd(Atomic& t, int incr)  { __exchange_and_add(&t, incr); return t; }

class CriticalSection {
protected:
	pthread_mutex_t  mutex[1];

	CriticalSection(int)         {}

public:
	void  Enter()             { pthread_mutex_lock(mutex); }
	void  Leave()             { pthread_mutex_unlock(mutex); }

	struct Lock;
	struct LockMain;

	CriticalSection()         { pthread_mutex_init(mutex, NULL); }
	~CriticalSection()        { pthread_mutex_destroy(mutex); }
};

#endif

CriticalSection& MainCriticalSection();

struct CriticalSection::Lock {
	CriticalSection& s;
	Lock(CriticalSection& s) : s(s) { s.Enter(); }
	~Lock()                         { s.Leave(); }
};

struct CriticalSection::LockMain : public CriticalSection::Lock {
	LockMain() : Lock(MainCriticalSection()) {}
};

class StaticCriticalSection {
	CriticalSection *section;
	
	void Initialize();

public:
	CriticalSection& Get()         { if(!section) Initialize(); return *section; }
	operator CriticalSection&()    { return Get(); }
	void Enter()                   { Get().Enter(); }
	void Leave()                   { Get().Leave(); }
};

#else // Not _MT

typedef int Atomic;

inline int  AtomicRead(const Atomic& t)      { return t; }
inline void AtomicWrite(Atomic& t, int data) { t = data; }

inline int  AtomicInc(Atomic& t)             { ++t; return t; }
inline int  AtomicDec(Atomic& t)             { --t; return t; }
inline int  AtomicXAdd(Atomic& t, int incr)  { Atomic x = t; t += incr; return t; }

class CriticalSection {
protected:
	friend CriticalSection& MainCriticalSection();

public:
	void  Enter()                {}
	void  Leave()                {}

	CriticalSection()            {}
	~CriticalSection()           {}

	struct Lock;
	struct LockMain;
};

typedef CriticalSection StaticCriticalSection;

struct CriticalSection::Lock {
	Lock(CriticalSection&) {}
	~Lock()                {}
};

struct CriticalSection::LockMain {
	LockMain() {}
};

#endif
