#include "Core.h"
#pragma hdrstop

VectorMap<Uuid, PteBase::PrecEx>& PteBase::Map()
{
	return Single<VectorMap<Uuid, PrecEx> >();
}

int PteBase::Find(const Uuid& uuid)
{
	return Map().Find(uuid);
}

void PteBase::Remove(int n)
{
	Map().Unlink(n);
}

int PteBase::Add(const Uuid& uuid) {
	return Map().Put(uuid, PrecEx());
}

PteBase::Prec *PteBase::PtrAdd()
{
	if(prec)
		prec->Retain();
	else {
		prec = new Prec;
		prec->n = EXTL + 1;
		prec->ptr = this;
	}
	return prec;
}

void PteBase::PtrRelease(Prec *prec)
{
	if(!prec) return;
	if(prec->n < EXTL) {
		if(--prec->Ex().n == 0 && prec->ptr == NULL) {
			Remove(prec->n);
			delete prec;
		}
	}
	else
	if(--prec->n == EXTL) {
		if(prec->ptr) prec->ptr->prec = NULL;
		delete prec;
	}
}

void PteBase::AddUuid(Prec *prec, const Uuid& uuid) {
	int n = Add(uuid);
	PrecEx& ex = Map(n);
	ex.n = prec->n - EXTL;
	ex.prec = prec;
	prec->n = n;
}

PteBase::Prec *PteBase::AddUuid(const Uuid& uuid) {
	Prec *prec = new Prec;
	prec->ptr = NULL;
	prec->n = 0;
	AddUuid(prec, uuid);
	return prec;
}

const Uuid& PteBase::GetUuid(Prec *prec) {
	if(prec->n >= EXTL)
		AddUuid(prec, Uuid::Create());
	return Map().GetKey(prec->n);
}

const Uuid& PteBase::GetUuid() {
	if(!prec) {
		prec = AddUuid(Uuid::Create());
		return Map().GetKey(prec->n);
	}
	return GetUuid(prec);
}

void PteBase::Free() {
	if(prec) {
		prec->ptr = NULL;
		if(prec->n < EXTL && prec->Ex().n == 0) {
			Remove(prec->n);
			delete prec;
		}
	}
}

PteBase::PteBase() {
	prec = NULL;
	PteBase::Map();
}

PteBase::~PteBase() {
	Free();
}

void PteBase::SetUuid(const Uuid& uuid)
{
	if(HasUuid())
		Map().SetKey(prec->n, uuid);
	else {
		int i = Find(uuid);
		if(i >= 0) {
			ASSERT(!prec);
			prec = Map(i).prec;
			ASSERT(prec->ptr == 0);
			prec->ptr = this;
		}
		else
		if(prec)
			AddUuid(prec, uuid);
		else {
			prec = AddUuid(uuid);
			prec->ptr = this;
		}
	}
}

void PteBase::SerializePte(Stream& s) {
	Uuid uuid;
	if(s.IsStoring())
		uuid = GetUuid();
	s % uuid;
	if(s.IsLoading())
		SetUuid(uuid);
}

Uuid PtrBase::GetUuid()
{
	Uuid uuid;
	if(prec)
		uuid = PteBase::GetUuid(prec);
	else
		uuid.SetNull();
	return uuid;
}

void PtrBase::SetUuid(const Uuid& uuid)
{
	if(uuid.IsNull())
		prec = NULL;
	else {
		PteBase::PtrRelease(prec);
		int i = PteBase::Find(uuid);
		prec = i >= 0 ? PteBase::Map(i).prec : PteBase::AddUuid(uuid);
		prec->Retain();
	}
}

void PtrBase::Serialize(Stream& s)
{
	Uuid uuid;
	if(s.IsStoring())
		uuid = GetUuid();
	s % uuid;
	if(s.IsLoading())
		SetUuid(uuid);
}


void PtrBase::Set(PteBase *p) {
	prec = p ? p->PtrAdd() : NULL;
}

PtrBase::PtrBase() {
	PteBase::Map();
}

PtrBase::~PtrBase() {
	Release();
}
