template <class T> class Ptr;

class PteBase {
protected:
	enum { EXTL = 0x80000000 };
	struct Prec;
	struct PrecEx : Moveable<PrecEx> {
		dword   n;
		Prec   *prec;
	};

	friend struct Prec;

	static  VectorMap<Uuid, PrecEx>& Map();
	static  PrecEx& Map(int i)                  { return Map()[i]; }
	static  int     Find(const Uuid& uuid);
	static  int     Add(const Uuid& uuid);
	static  void    Remove(int n);

	struct Prec {
		PteBase *ptr;
		dword    n;
		PrecEx&  Ex()     { ASSERT(n < EXTL); return Map(n); }
		void     Retain() { if(n < EXTL) Ex().n++; else n++; }
	};
	Prec         *prec;

	Prec         *PtrAdd();
	static void   PtrRelease(Prec *prec);
	static Prec  *PtrAdd(const Uuid& uuid);

	static void  AddUuid(Prec *prec, const Uuid& uuid);
	static Prec *AddUuid(const Uuid& uuid);
	static const Uuid& GetUuid(Prec *prec);

	PteBase();
	~PteBase();

	void   Free();

	friend class PtrBase;

public:
	bool         HasUuid()              { return prec && prec->n < EXTL; }
	const Uuid&  GetUuid();
	void         SetUuid(const Uuid& uuid);

	void         SerializePte(Stream& s);
};

class PtrBase {
protected:
	PteBase::Prec *prec;
	void Set(PteBase *p);
	void Release()                            { PteBase::PtrRelease(prec); }

public:
	bool  HasUuid()                           { return prec && prec->n < PteBase::EXTL; }
	Uuid  GetUuid();
	void  SetUuid(const Uuid& uuid);

	void  Serialize(Stream& s);

	PtrBase();
	~PtrBase();
};

template <class T>
class Pte : public PteBase {
	friend class Ptr<T>;
};

template <class T>
class Ptr : public PtrBase, Moveable< Ptr<T> > {
	T   *Get() const                          { return prec ? static_cast<T *>(prec->ptr) : NULL; }

public:
	T       *operator->() const               { return Get(); }
	T       *operator~() const                { return Get(); }
	operator T*() const                       { return Get(); }

	Ptr& operator=(T *ptr)                    { Release(); Set(ptr); return *this; }
	Ptr& operator=(const Ptr& ptr)            { Release(); Set(ptr.Get()); return *this; }

	Ptr()                                     { prec = NULL; }
	Ptr(T *ptr)                               { Set(ptr); }
	Ptr(const Ptr& ptr)                       { Set(ptr.Get()); }

	String ToString() const;

	friend bool operator==(const Ptr& a, const T *b)   { return a.Get() == b; }
	friend bool operator==(const T *a, const Ptr& b)   { return a == b.Get(); }
	friend bool operator==(const Ptr& a, const Ptr& b) { return a.prec == b.prec; }
//G++ needs this...
	friend bool operator==(const Ptr& a, T *b)         { return a.Get() == b; }
	friend bool operator==(T *a, const Ptr& b)         { return a == b.Get(); }

	friend bool operator!=(const Ptr& a, const T *b)   { return a.Get() != b; }
	friend bool operator!=(const T *a, const Ptr& b)   { return a != b.Get(); }
	friend bool operator!=(const Ptr& a, const Ptr& b) { return a.prec != b.prec; }
//G++ needs this...
	friend bool operator!=(const Ptr& a, T *b)         { return a.Get() != b; }
	friend bool operator!=(T *a, const Ptr& b)         { return a != b.Get(); }
};

template <class T>
String Ptr<T>::ToString() const
{
	return prec ?
			prec->ptr && prec->ptr->HasUuid() ?
				::FormatPtr(Get()) + " " + ::AsString(prec->ptr->GetUuid())
			    : ::FormatPtr(Get())
		    : String("0x0");
}
