#include "CppBase.h"

#define LLOG(x)
#define LTIMING(x)

String Qualify(const CppBase& base, const Vector<String>& nd, const String& type)
{
	if(IsNull(type) || type[0] == ':')
		return type;
	for(int i = 0; i < nd.GetCount(); i++) {
		String q = nd[i] + type;
		if(base.Find(q) >= 0)
			return q;
	}
	return type;
}

Vector<String> MakeNests(String nn)
{
	Vector<String> nests;
	while(nn.GetCount()) {
		nests.Add(nn + "::");
		int q = nn.ReverseFind(':');
		nn.Trim(max(0, q - 1));
	}
	nests.Add("::");
	return nests;
}

String QualifyIds(const CppBase& base, const Vector<String>& nests, const String& k, bool ignore = false)
{
	String r;
	CParser p(k);
	while(!p.IsEof()) {
		if(p.Char2(':', ':'))
			while(p.Char2(':', ':')) {
				r << "::";
				if(p.IsId())
					r << p.ReadId();
			}
		else
		if(p.IsId()) {
			String t = p.ReadId();
			while(p.Char2(':', ':')) {
				t << "::";
				if(p.IsId())
					t << p.ReadId();
			}
			if(!ignore)
				t = Qualify(base, nests, t);
			if(iscid(*r.Last()) && iscid(*t))
				r << ' ';
			r << t;
		}
		else {
			int c = p.GetChar();
			if(c != '~')
				ignore = false;
			r.Cat(c);
			p.Spaces();
		}
	}
	return r;
}

String Qualify(const CppBase& base, const String& nest, const String& type)
{
	return QualifyIds(base, MakeNests(nest), type);
}

String QualifyKey(const CppBase& base, const String& nest, const String& type)
{
	return QualifyIds(base, MakeNests(nest), type, true);
}

void QualifyTypes(CppBase& base, const Vector<String>& nests, CppItem& m)
{
	m.qtype = QualifyIds(base, nests, m.type);
	m.qptype = QualifyIds(base, nests, m.ptype);
}

void QualifyTypes(CppBase& base, const String& nest, CppItem& m)
{
	QualifyTypes(base, MakeNests(nest), m);
}

void Qualify(CppBase& base, bool force)
{
	LTIMING("Qualify");
	for(int i = 0; i < base.GetCount(); i++) {
		CppNest& n = base[i];
		Vector<String> nests = MakeNests(base.GetKey(i));
		Vector<int> rm;
		for(int i = 0; i < n.GetCount(); i++) {
			CppItem& m = n.item[i];
			if(!m.qualified || force) {
				m.qualified = true;
				QualifyTypes(base, nests, m);
				if(m.IsCode()) {
					String k = n.key[i];
					String r = QualifyIds(base, nests, k, true);
					if(k != r) {
						int q = n.key.Find(r);
						if(q >= 0 && q != i)
							if(m.decla) {
								m.pos.Append(n.item[q].pos);
								rm.Add(q);
								n.key.Set(i, r);
							}
							else {
								n.item[q].pos.Append(m.pos);
								rm.Add(i);
							}
						else
							n.key.Set(i, r);
					}
				}
			}
		}
		Sort(rm);
		n.Remove(rm);
	}
}

void Remove(CppBase& base, const Vector<String>& pf)
{
	int ni = 0;
	Vector<int> file;
	for(int i = 0; i < pf.GetCount(); i++)
		file.Add(GetCppFileIndex(pf[i]));
	while(ni < base.GetCount()) {
		CppNest& n = base[ni];
		Vector<int> nr;
		for(int mi = 0; mi < n.GetCount(); mi++) {
			CppItem& m = n.item[mi];
			int i = 0;
			while(i < m.pos.GetCount())
				if(FindIndex(file, m.pos[i].file) >= 0)
					m.pos.Remove(i);
				else
					i++;
			if(m.pos.GetCount() == 0)
				nr.Add(mi);
		}
		n.Remove(nr);
		if(n.GetCount() == 0)
			base.Remove(ni);
		else
			ni++;
	}
}

void CppSimpleItem::Serialize(Stream& s)
{
	s % kind % access;
	s % natural % at % tparam % param % pname % tname % type % ptype % virt;
}
