#include "CtrlLib.h"

void QtfCtrl::MouseWheel(Point p, int zdelta, dword keyflags) {
	sb.Wheel(zdelta);
}

QtfCtrl::QtfCtrl() {
	zoom = VertLayoutZoom(140);
	sb.SetLine(8).Hide();
	sb.WhenScroll = callback(this, &QtfCtrl::Scroll);
	NoWantFocus();
	tm = bm = lm = rm = 0;
	vcenter = false;
	color = SLtGray;
	Background(Null);
	AddFrame(sb);
}

QtfCtrl::~QtfCtrl() {}

QtfCtrl& QtfCtrl::SetLabel(const char *text) {
	Set(text);
	return *this;
}

bool QtfCtrl::Key(dword key, int count) {
	return sb.VertKey(key);
}

int QtfCtrl::GetHeight(int cx) const {
	cx -= lm + rm;
	if(cx <= 0) return 0;
	return doc.GetHeight(zoom, ScreenInfo(), cx, cache);
}

void QtfCtrl::Layout() {
	Size sz = sb.GetViewSize();
	if(GetHeight(sz.cx) > sz.cy - tm - bm) {
		sb.Show();
		sb.Set(sb, GetSize().cy - tm - bm, GetHeight(sb.GetReducedViewSize().cx));
	}
	else
		sb.Hide();
//	if(WhenLink) //!! fails when text is set prior to WhenLink
	{
		int cx = GetSize().cx - lm - rm;
		if(cx <= 0) return;
		valuerect = doc.GetValueRects(zoom, ScreenInfo(), 0, 0, cx);
	}
}

Value QtfCtrl::FindValue(Point p) const {
	p.y += sb - tm;
	p.x -= lm;
	for(int i = 0; i < valuerect.GetCount(); i++) {
		const ValueRect& r = valuerect[i];
		if(r.rect.Contains(p))
			return r.value;
	}
	return Null;
}

Image QtfCtrl::CursorImage(Point p, dword) {
	if(!WhenLink) return Image::Arrow();
	Value v = FindValue(p);
	return IsNull(v) ? Image::Arrow() : CtrlImg::HandCursor();
}

void  QtfCtrl::LeftUp(Point p, dword keyflags) {
	SetWantFocus();
	if(!WhenLink) return;
	Value v = FindValue(p);
	if(IsNull(v)) return;
	WhenLink(v);
}

void QtfCtrl::Paint(Draw& w) {
	Size sz = GetSize();
	w.DrawRect(0, 0, sz.cx, sz.cy, color);
	int hg = GetHeight(sz.cx);
	sz.cx -= lm + rm;
	sz.cy -= tm + bm;
	w.Clipoff(lm, tm, sz.cx, sz.cy);
	int dcy = vcenter && hg <= sz.cy - tm - bm ? (sz.cy - hg) : 0;
	doc.Paint(zoom, w, 0, dcy / 2, sz.cx, sz.cy - dcy, sb, cache, color);
	w.End();
}

void QtfCtrl::Scroll() {
	Refresh();
}

void QtfCtrl::Set(const char *s) {
	cache.Clear();
	doc.Clear();
	doc.Qtf(s, GetDefaultCharset());
	sb = 0;
	Layout();
	Refresh();
}

void QtfCtrl::Set(const Document& _doc) {
	cache.Clear();
	doc <<= _doc;
	sb = 0;
	Layout();
	Refresh();
}

void QtfCtrl::Pick(pick_ Document& _doc) {
	cache.Clear();
	doc = _doc;
	sb = 0;
	Layout();
	Refresh();
}

void QtfPopUp::Deactivate() {
	if(open)
		DoClose();
}

void QtfPopUp::CancelMode() {
	DoClose();
}

void QtfPopUp::DoClose() {
	if(open) {
		open = false;
		Close();
	}
}

void QtfPopUp::PopUp(Ctrl *parent) {
	Close();
	Rect r = Rect(0, 0, width, maxheight);
	GetFrame().FrameLayout(r);
	int cy = min(maxheight, GetHeight(r.Width()) + tm + bm + maxheight - r.Height());
	Rect area = GetWorkArea();
	Point p = GetMousePos();
	r.top = max(area.top, p.y + 16);
	r.bottom = r.top + cy;
	if(r.bottom > area.bottom) {
		r.bottom = area.bottom;
		r.top = r.bottom - cy;
	}
	r.left = max(area.left, p.x - width / 2);
	r.right = r.left + width;
	if(r.right > area.right) {
		r.right = area.right;
		r.left = r.right - width;
	}
	open = false;
	SetRect(r);
	Ctrl::PopUp(parent);
	SetFocus();
	open = true;
}

QtfPopUp::QtfPopUp() {
	static ColorF helpborder[] = {
		(ColorF)2,
		&SBlack, &SBlack, &SBlack, &SBlack,
		&SLtGray, &SLtGray, &SLtGray, &SLtGray,
	};
	static BorderFrame helpframe(helpborder);
	width = 250;
	maxheight = 150;
	Margins(5).Background(LtYellow).SetFrame(helpframe);
	open = false;
}
