#include "CtrlLib.h"

#define LLOG(x)   // LOG(x)

ToolTip::ToolTip()
{
	SetFrame(BlackFrame());
}

Size ToolTip::GetMinSize() const
{
	return AddFrameSize(ScreenInfo().GetTextSize(text) + 4);
}

void ToolTip::Paint(Draw& w)
{
	Size sz = GetSize();
	w.DrawRect(sz, SLtYellow);
	w.DrawText(2, 2 + sz.cy - ScreenInfo().GetTextSize(text).cy - 4, text);
}

void ToolTip::PopUp(Ctrl *owner, Point p, bool effect)
{
	LLOG("ToolTip::PopUp" << Desc(owner) << " @ " << p);
	Rect r = GetWorkArea();
	Size sz = GetMinSize();
	p.x = max(p.x + sz.cx > r.right ? r.right - sz.cx : p.x, 0);
	p.y = max(p.y + sz.cy > r.bottom ? r.bottom - sz.cy : p.y, 0);
	if(IsFlag(EFFECT_SLIDE) && effect)
		SetRect(p.x, p.y, sz.cx, 1);
	else
		SetRect(p.x, p.y, sz.cx, sz.cy);
	Ctrl::PopUp(owner, true, false, true);
	if(effect)
		Animate(*this, p.x, p.y, sz.cx, sz.cy);
}

Ptr<Ctrl> tipctrl;
bool      showmode;

ToolTip& AppToolTip()
{
	static ToolTip tip;
	return tip;
}

void CloseToolTip()
{
	ToolTip& q = AppToolTip();
	if(q.IsOpen())
		q.Close();
}

void EndShowMode()
{
	showmode = false;
}

void ShowToolTip()
{
	LLOG("ShowToolTip");
	CloseToolTip();
	if(tipctrl) {
		String text = tipctrl->GetTip();
		LLOG("-> showing tip: " << text);
		Ctrl *top = tipctrl->GetTopWindow();
		if(!text.IsEmpty() && top && top->IsForeground()) {
			LLOG("-> foreground");
			ToolTip& q = AppToolTip();
			q.Set(text);
			q.PopUp(top, GetMousePos() + Size(0, 22), !showmode);
			showmode = true;
			KillTimeCallback((void *)EndShowMode);
			return;
		}
		LLOG("-> background / empty text, top = " << ::Name(top));
	}
	SetTimeCallback(200, callback(EndShowMode), (void *)EndShowMode);
}

bool ToolTipHook(Ctrl *ctrl, bool inframe, int event, Point p, int zdelta, dword keyflags)
{
	if(event == Ctrl::MOUSEMOVE && ctrl != &AppToolTip())
	{
		ctrl = Ctrl::GetVisibleChild(ctrl, p, inframe);
		if(ctrl != tipctrl) {
			LLOG("ToolTipHook / ctrl change -> " << ::Name(ctrl));
			tipctrl = ctrl;
			KillTimeCallback((void *)ToolTipHook);
			if(showmode)
				ShowToolTip();
			else {
				LLOG("ToolTipHook -> SetTimeCallback");
				SetTimeCallback(1000, callback(ShowToolTip), (void *)ToolTipHook);
			}
		}
	}
	if(event == Ctrl::MOUSELEAVE) {
		CloseToolTip();
		SetTimeCallback(1000, callback(EndShowMode), (void *)EndShowMode);
	}
	if((event & Ctrl::ACTION) == Ctrl::DOWN) {
		CloseToolTip();
		EndShowMode();
	}
	return false;
}

bool ToolTipStateHook(Ctrl *ctrl, int reason)
{
	if((reason == Ctrl::CLOSE || reason == Ctrl::ENABLE) && ctrl != &AppToolTip()) {
		LLOG("ToolTipStateHook -> close");
		CloseToolTip();
		EndShowMode();
	}
	return false;
}

INITBLOCK {
	LLOG("Ctrl::InstallMouseHook(ToolTipHook)");
	Ctrl::InstallMouseHook(ToolTipHook);
	Ctrl::InstallStateHook(ToolTipStateHook);
}

void LinkToolTipIn__() {}  // "Call" from elsewhere to force linking this module in...

void PerformDescription()
{
	Ctrl *top = Ctrl::GetActiveWindow();
	if(!top) return;
	Vector<Image> vi;
	vi.Add(CtrlImg::HelpCursor0());
	vi.Add(CtrlImg::HelpCursor1());
	if(PointLoop(*top, vi, 300)) {
		Point p = GetMousePos();
		Ctrl *ctrl = Ctrl::GetVisibleChild(top, p - top->GetRect().TopLeft(), true);
		while(ctrl) {
			String desc = ctrl->GetDescription();
			if(!IsEmpty(desc)) {
				static QtfPopUp qtf;
				qtf.Set(desc);
				qtf.PopUp(top);
				return;
			}
			ctrl = ctrl->GetParent();
		}
	}
}
