#include "ide.h"

#define LDUMP(x)   // DUMP(x)
#define LDUMPC(x)  // DUMPC(x)
#define LLOG(x)    // LOG(x)

AssistEditor::AssistEditor()
{
	assist.Header();
	assist.NoGrid();
	assist.AddColumn().Margin(0).SetDisplay(Single<CppItemInfoDisplay>());
	assist.WhenLeftClick = THISBACK(AssistInsert);
	auto_assist = true;
}

Vector<String> TemplatePars(const String& type)
{
	Vector<String> r;
	int q = type.Find('<');
	if(q >= 0) {
		const char *s = ~type + q + 1;
		int lvl = 0;
		String t;
		while(*s && (lvl || *s != '>')) {
			if(*s == '<')
				lvl++;
			if(*s == '>')
				lvl--;
			if(*s == ',') {
				r.Add(t);
				t.Clear();
				s++;
			}
			else
				t.Cat(*s++);
		}
		r.Add(t);
	}
	return r;
}

void SubstituteTpars(Vector<String>& type, const String& tname)
{
	Vector<String> tp = TemplatePars(tname);
	for(int i = 0; i < type.GetCount(); i++)
		if(IsDigit(type[i][0])) {
			int q = atoi(type[i]);
			if(q >= 0 && q < tp.GetCount()) {
				LLOG(type[i] << " -> " << tp[q]);
				type[i] = tp[q];
			}
		}
}

bool AssistEditor::NestId(const CppNest& n, const String& id, Vector<String>& type, bool& code, String& t)
{
	int q = n.name.Find(NoTemplatePars(id));
	if(q >= 0) {
		do {
			const CppItem& m = n[q];
			const String& qt = m.qtype;
			if(!IsNull(qt) && FindIndex(type, qt) < 0 && (m.IsCode() || m.IsData())) {
				type.Add(qt);
				if(m.IsCode())
					code = true;
			}
			q = n.name.FindNext(q);
		}
		while(q >= 0);
		return true;
	}
	q = n.key.Find("class");
	if(q < 0)
		q = n.key.Find("struct");
	LDUMP(q);
	if(q >= 0) {
		Vector<String> base = Split(n[q].qptype, ';');
		LDUMP(id);
		LDUMPC(base);
		SubstituteTpars(base, t);
		for(int i = 0; i < base.GetCount(); i++) {
			q = BrowserBase().Find(NoTemplatePars(base[i]));
			String h = base[i];
			if(q >= 0 && NestId(BrowserBase()[q], id, type, code, h)) {
				t = h;
				return true;
			}
		}
	}
	return false;
}

void AssistScanError(int line, const String& text)
{
	PutVerbose(String().Cat() << "(" << line << "): " << text);
}

void AssistEditor::Context(Parser& parser, int pos)
{
	theide->ScanFile();
	String s = Get(0, pos);
	StringStream ss(s);
	parser.dobody = true;
	parser.Do(ss, IgnoreList(), BrowserBase(), Null, callback(AssistScanError));
	QualifyTypes(BrowserBase(), parser.current_nest, parser.current);
}

void AssistEditor::TypeOf(const String& id, Vector<String>& r, bool& code)
{
	Parser parser;
	Context(parser, GetCursor());
	code = false;
	if(IsNull(id)) {
		r.Add(parser.current_nest);
		if(parser.current_nest != "::")
			r.Add("::");
		return;
	}
	int q = parser.local.FindLast(id);
	if(q >= 0) {
		r.Add(parser.local[q]);
		return;
	}
	Vector<String> p = Split(parser.current.pname, ';', false);
	q = FindIndex(p, id);
	if(q >= 0) {
		r.Add(Split(parser.current.qptype, ';', false)[q]);
		return;
	}
	q = BrowserBase().Find(parser.current_nest);
	String dummy;
	if(q >= 0 && NestId(BrowserBase()[q], id, r, code, dummy))
		return;
	q = BrowserBase().Find("::"); // Better namespace support needed !!!
	if(q >= 0)
		NestId(BrowserBase()[q], id, r, code, dummy);
}

Vector<String> AssistEditor::Operator(const char *oper, const Vector<String>& type)
{
	Vector<String> nt;
	bool d;
	for(int i = 0; i < type.GetCount(); i++) {
		String t = type[i];
		int q = BrowserBase().Find(NoTemplatePars(t));
		if(q < 0 || !NestId(BrowserBase()[q], oper, nt, d, t))
			nt.Add(type[i]);
		SubstituteTpars(nt, t);
	}
	return nt;
}

Vector<String> AssistEditor::TypeOf(const Vector<String>& xp)
{
	Vector<String> type;
	if(xp.GetCount() == 0 || !iscid(xp[0][0]))
		return type;
	bool code;
	TypeOf(xp[0], type, code);
	LDUMPC(xp);
	LDUMPC(type);
	int q = 1;
	if(code && xp.GetCount() > 1 && xp[1][0] == '(')
		q++;
	while(q < xp.GetCount() && type.GetCount()) {
		int c = xp[q][0];
		LDUMP(xp[q]);
		if(iscid(c)) {
			Vector<String> nt;
			bool code = false;
			for(int i = 0; i < type.GetCount(); i++) {
				String t = type[i];
				int j = BrowserBase().Find(NoTemplatePars(t));
				if(j >= 0)
					NestId(BrowserBase()[j], xp[q], nt, code, t);
				LDUMPC(nt);
				SubstituteTpars(nt, t);
			}
			if(code && q + 1 < xp.GetCount() && xp[q + 1][0] == '(')
				q++;
			type = nt;
		}
		if(c == '(')
			type = Operator("operator()", type);
		if(c == '[')
			type = Operator("operator[]", type);
		if(c == '-')
			type = Operator("operator->", type);
		q++;
	}
	LDUMPC(type);
	return type;
}

int CharFilterT(int c)
{
	return c >= '0' && c <= '9' ? "TUVWXYZMNO"[c - '0'] : c;
}

void GatherItems(const String& type, Array<CppItemInfo>& item, const Index<String>& hidden, String& header, bool nom, Index<String>& in_types)
{
	LLOG("GatherItems " << type);
	if(in_types.Find(type) >= 0) {
		LLOG("-> recursion, exiting");
		return;
	}
	in_types.Add(type);
	if(!IsNull(header))
		header << ", ";
	header << Filter(type, CharFilterT);
	int q = BrowserBase().Find(NoTemplatePars(type));
	LDUMP(q);
	if(q < 0)
		return;
	CppNest& m = BrowserBase()[q];
	Index<String> names;
	String base;
	for(int i = 0; i < m.GetCount(); i++) {
		const CppItem& im = m[i];
		String name = m.name[i];
		if(im.IsType())
			base = im.qptype;
		LDUMP(name);
		if(hidden.Find(name) < 0 && (im.IsCode() || im.IsData()) && (nom || im.access == PUBLIC)) {
			names.FindAdd(name);
			CppItemInfo& f = item.Add();
			f.name = name;
			(CppSimpleItem&)f = im;
		}
	}
	Vector<String> b = Split(base, ';');
	SubstituteTpars(b, type);
	for(int i = 0; i < b.GetCount(); i++)
		GatherItems(b[i], item, names, header, nom, in_types);
	in_types.Drop();
}

struct CppItemInfoOrder {
	bool operator()(const CppItemInfo& a, const CppItemInfo& b) const {
		return CombineCompare(a.name, b.name)(a.natural, b.natural) < 0;
	}
};

void AssistEditor::CloseAssist()
{
	if(assist.IsOpen())
		assist.Close();
	assist_item.Clear();
}

String AssistEditor::ReadIdBack(int q)
{
	String id;
	while(q > 0 && iscid(GetChar(q - 1)))
		q--;
	while(q < GetLength() && iscid(GetChar(q)))
		id << (char)GetChar(q++);
	return id;
}

int AssistEditor::Ch(int i)
{
	return i >= 0 && i < GetLength() ? GetChar(i) : 0;
}

int AssistEditor::ParsBack(int q)
{
	int level = 1;
	--q;
	while(q > 0) {
		if(Ch(q) == ']' || Ch(q) == ')' || Ch(q) == '}')
			level++;
		if(Ch(q) == '[' || Ch(q) == '(' || Ch(q) == '{')
			if(--level <= 0)
				break;
		--q;
	}
	return max(0, q);
}

bool IsSpc(int c)
{
	return c > 0 && c <= 32;
}

Vector<String> AssistEditor::ReadBack(int q)
{
	Vector<String> r;
	for(;;) {
		if(r.GetCount() > 100) {
			r.Clear();
			break;
		}
		while(q > 0 && IsSpc(Ch(q - 1)))
			q--;
		int c = Ch(q - 1);
		int c1 = Ch(q - 2);
		if(iscid(c)) {
			while(iscid(Ch(q - 1)))
				q--;
			if(!iscib(Ch(q)))
				break;
			r.Add();
			int i = q;
			while(i < GetLength() && iscid(Ch(i)))
				r.Top() << (char)Ch(i++);
		}
		else
		if(c == ']') {
			r.Add("[]");
			q = ParsBack(q - 1);
			continue;
		}
		else
		if(c == ')') {
			r.Add("()");
			q = ParsBack(q - 1);
			continue;
		}
		c = Ch(q - 1);
		if(c == '>' && Ch(q - 2) == '-') {
			r.Add("->");
			q -= 2;
			continue;
		}
		if(c == '.') {
			r.Add(".");
			q--;
			continue;
		}
		break;
	}
	Reverse(r);
	return r;
}

int AssistEditor::SyncAssist(bool readid)
{
	String name;
	if(readid)
		name = ReadIdBack(GetCursor());
	int cx = 0;
	assist.Clear();
	for(int i = 0; i < assist_item.GetCount(); i++) {
		const CppItemInfo& m = assist_item[i];
		if(memcmp(name, m.name, name.GetCount()) == 0) {
			Value v = RawToValue(m);
			cx = max(cx, Single<CppItemInfoDisplay>().GetStdSize(v).cx);
			assist.Add(v);
		}
	}
	return cx;
}

void AssistEditor::Assist()
{
	if(!assist_active)
		return;
	StartBrowserBase();
	CloseAssist();
	int q = GetCursor();
	assist_cursor = q;
	while(iscid(Ch(q - 1)))
		q--;
	Vector<String> h = ReadBack(q);
	Vector<String> type;
	bool d;
	if(h.GetCount() == 0)
		TypeOf(Null, type, d);
	else {
		type = TypeOf(h);
		if(type.GetCount() == 0)
			return;
	}
	Index<String> hidden;
	String header;
	for(int i = 0; i < type.GetCount(); i++) {
		Index<String> in_types;
		GatherItems(type[i], assist_item, hidden, header, h.GetCount() == 0, in_types);
	}
	if(assist_item.GetCount() == 0)
		return;
	Sort(assist_item, CppItemInfoOrder());
	int lcy = ScreenInfo().GetFontInfo(Arial(11)).GetHeight() + 3;
	Size sz;
	sz.cx = min(600, max(SyncAssist(false), ScreenInfo().GetTextSize(header).cx + 2 * HeaderCtrl::GetStdHeight()));
	sz.cy = min(300, lcy * assist.GetCount());
	assist.HeaderTab(0).SetText(header);
	sz.cx += 4 + ScrollBar::GetStdBox() + 12;
	sz.cy += 4;
	sz.cy += HeaderCtrl::GetStdHeight();
	assist.SetLineCy(lcy);
	Point p = GetCaretPoint() + GetScreenView().TopLeft();
	Rect r(p, sz);
	Rect wa = GetWorkArea();
	if(p.x + sz.cx > wa.right)
		p.x = wa.right - sz.cx;
	if(p.y + sz.cy + GetFontSize().cy < wa.bottom)
		assist.SetRect(p.x, p.y + GetFontSize().cy, sz.cx, sz.cy);
	else
		assist.SetRect(p.x, p.y - sz.cy, sz.cx, sz.cy);
	assist.Ctrl::PopUp(this, false, false, true);
	SyncAssist();
}

void AssistEditor::AssistInsert()
{
	if(assist.IsCursor()) {
		const CppItemInfo& f = ValueTo<CppItemInfo>(assist.Get(0));
		String txt = f.name;
		int l = txt.GetCount();
		int pl = txt.GetCount();
		if(f.kind >= FUNCTION && f.kind <= INLINEFRIEND) {
			txt << '(';
			pl = l = txt.GetCount();
			if(IsNull(f.param))
				l++;
			else {
				Vector<String> p = Split(f.param, ';', false);
				for(int i = 0; i < p.GetCount(); i++) {
					if(i) {
						txt << (char)184;
						txt << ' ';
					}
					txt << p[i];
					if(i == 0)
						pl = txt.GetCount();
				}
			}
			txt << ')';
		}
		int cl = GetCursor();
		int ch = cl;
		while(iscid(Ch(cl - 1)))
			cl--;
		while(iscid(Ch(ch)))
			ch++;
		Remove(cl, ch - cl);
		SetCursor(cl);
		Paste(ToUnicode(txt, CHARSET_WIN1250));
		if(pl > l)
			SetSelection(cl + l, cl + pl);
		else
			SetCursor(cl + l);
	}
	CloseAssist();
	IgnoreMouseUp();
}

bool AssistEditor::Key(dword key, int count)
{
	if(assist.IsOpen()) {
		if(key == K_UP || key == K_PAGEUP || key == K_CTRL_PAGEUP || key == K_CTRL_END)
			if(assist.IsCursor())
				return assist.Key(key, count);
			else {
				assist.SetCursor(assist.GetCount() - 1);
				return true;
			}
		if(key == K_DOWN || key == K_PAGEDOWN || key == K_CTRL_PAGEDOWN || key == K_CTRL_HOME)
			if(assist.IsCursor())
				return assist.Key(key, count);
			else {
				assist.SetCursor(0);
				return true;
			}
		if(key == K_ENTER && assist.IsCursor()) {
			AssistInsert();
			return true;
		}
	}
#ifdef _DEBUG
	if(key == K_F12) {
		PutConsole("Type is");
		Vector<String> h = ReadBack(GetCursor());
		Vector<String> r = TypeOf(h);
		for(int i = 0; i < r.GetCount(); i++)
			PutConsole(r[i]);
	}
#endif
	int c = GetCursor();
	int cc = GetChar(c);
	int bcc = c > 0 ? GetChar(c - 1) : 0;
	bool b = CodeEditor::Key(key, count);
	if(assist.IsOpen()) {
		int c = GetCursor();
		if(!iscid(key) &&
		   !(iscid(cc) && (key == K_DELETE || key == K_RIGHT)) &&
		   !(iscid(bcc) && (key == K_LEFT || key == K_BACKSPACE))) {
			if(b) {
				CloseAssist();
				if(key == '.')
					Assist();
			}
		}
		else
			SyncAssist();
	}
	else {
		if(auto_assist && (key == '.' || key == '>' && Ch(GetCursor() - 2) == '-'))
			Assist();
	}
	if(key == ',') {
		if(Ch(GetCursor()) == 184) {
			int q = GetCursor() + 1;
			int lvl = 0;
			while(q < GetLength()) {
				int ch = Ch(q);
				if(ch == '\n')
					return b;
				if(ch == 184 || lvl <= 0 && ch == ')') {
					SetSelection(GetCursor(), q);
					return b;
				}
				if(ch == '(')
					lvl++;
				if(ch == '(')
					lvl--;
				q++;
			}
		}
	}
	return b;
}

void AssistEditor::MouseWheel(Point p, int zdelta, dword keyflags)
{
	if(assist.IsOpen())
		assist.MouseWheel(p, zdelta, keyflags);
	else
		CodeEditor::MouseWheel(p, zdelta, keyflags);
}

void AssistEditor::LeftDown(Point p, dword keyflags)
{
	CloseAssist();
	CodeEditor::LeftDown(p, keyflags);
}

void AssistEditor::LostFocus()
{
	CloseAssist();
}

String MakeDefinition(const String& cls, const String& _n)
{
	String n = TrimLeft(_n);
	CParser p(n);
	while(!p.IsEof()) {
		const char *b = p.GetPtr();
		if(p.Id("operator"))
			return NormalizeSpaces(String(~n, b) + ' ' + cls + "::" + b);
		if(p.IsId()) {
			String id = p.ReadId();
			if(p.Char('('))
				return NormalizeSpaces(String(~n, b) + ' ' + cls + "::" + id + '(' + p.GetPtr());
		}
		else
			p.SkipTerm();
	}
	return n;
}

void AssistEditor::DCopy()
{
	String r;
	int l, h;
	if(!GetSelection(l, h)) {
		int i = GetLine(cursor);
		l = GetPos(i);
		h = l + line[i].GetLength() + 1;
	}
	Parser ctx;
	Context(ctx, l);
	String txt = Get(l, h - l);
	StringStream ss(txt);
	Parser parser;
	CppBase cpp;
	parser.Do(ss, IgnoreList(), cpp, Null, CNULL);
	String cls;
	cls = ctx.current_nest.Mid(ctx.current_namespacel);
	if(cls[0] == ':' && cls[1] == ':')
		cls = cls.Mid(2);
	for(int i = 0; i < cpp.GetCount(); i++) {
		const CppNest& nest = cpp[i];
		for(int j = 0; j < nest.GetCount(); j++) {
			const CppItem& m = nest[j];
			if(m.IsCode())
				if(cls.GetCount())
					r << MakeDefinition(cls, m.natural) << "\n{\n}\n\n";
				else
					r << '\t' << m.natural << ";\n";
		}
	}
	WriteClipboardText(r);
}
