#include "Common.h"

String GetCommonDir()
{
	return GetVar("COMMON");
}

String CommonPath(const String& filename)
{
	String s = GetCommonDir();
	if(s.IsEmpty())
		return ConfigFile(filename);
	return AppendFileName(s, filename);
}

String GetLocalDir()
{
	return ConfigFile("UppLocal");
}

String LocalPath(const String& filename)
{
	return AppendFileName(GetLocalDir(), filename);
}

String FollowCygwinSymlink(const String& file) {
	for(String fn = file;;) {
		if(fn.IsEmpty())
			return fn;
		FileIn fi(fn);
		if(!fi.IsOpen())
			return fn;
		char buffer[10];
		if(!fi.GetAll(buffer, 10) || memcmp(buffer, "!<symlink>", 10))
			return fn;
		fn = NormalizePath(LoadStream(fi), GetFileDirectory(fn));
	}
}

const char *SemiTextTest::Accept(const char *s) const {
	if(*s != ';') return NULL;
	s++;
	while(*s == ' ') s++;
	return s;
};

Vector<String> SplitDirs(const char *s) {
	return Split(s, Single<SemiTextTest>());
}

static GLOBAL_VARP(String, varsname, ("default"));
typedef VectorMap<String, String> sHT;
static GLOBAL_VAR(sHT, sVar);

String GetVarsName() {
	return varsname();
}

String VarFilePath(String name) {
	return ConfigFile(name + ".var");
}

String VarFilePath() {
	return VarFilePath(GetVarsName());
}

bool SaveVarFile(const char *filename, const VectorMap<String, String>& var)
{
	FileOut out(filename);
	for(int i = 0; i < var.GetCount(); i++)
		out << var.GetKey(i) << " = " << AsCString(var[i]) << ";\n";
	out.Close();
	return !out.IsError();
}

bool SaveVars(const char *name)
{
	if(!SaveVarFile(VarFilePath(name), sVar()))
		return false;
	varsname() = name;
	return true;
}

bool LoadVarFile(const char *name, VectorMap<String, String>& _var)
{
	try {
		VectorMap<String, String> var;
		String env = LoadFile(name);
		CParser p(env);
		while(!p.IsEof()) {
			String v = p.ReadId();
			p.Char('=');
			if(p.IsString())
				var.GetAdd(v) = p.ReadString();
			else {
				String ln;
				while(p.PeekChar() != '\r' && p.PeekChar() != '\n' && p.PeekChar() != ';')
					ln.Cat(p.GetChar());
				var.GetAdd(v) = ln;
				p.Spaces();
			}
			p.Char(';');
		}
		_var = var;
		return true;
	}
	catch(...) {
		return false;
	}
}

void   LoadVars(const char *name) {
	varsname() = (name && *name ? name : "default");
	LoadVarFile(VarFilePath(), sVar());
}

String GetVar(const String& var) {
	return sVar().Get(var, Null);
}

void SetVar(const String& var, const String& val) {
	sVar().GetAdd(var) = val;
	SaveVars(GetVarsName());
}

Vector<String> GetUppDirs() {
	return SplitDirs(GetVar("UPP"));
}

String GetUppDir() {
	Vector<String> s = GetUppDirs();
#ifdef PLATFORM_WIN32
	return s.GetCount() == 0 ? GetFileFolder(GetExeFilePath()) : s[0];
#endif
#ifdef PLATFORM_POSIX
	return s.GetCount() == 0 ? GetHomeDirectory() : s[0];
#endif
}

bool IsCSourceFile(const char *path)
{
	String ext = ToLower(GetFileExt(path));
	return ext == ".cpp" || ext == ".c" || ext == ".cc" || ext == ".cxx" || ext == ".icpp";
}

bool IsCHeaderFile(const char *path)
{
	String ext = ToLower(GetFileExt(path));
	return ext == ".h" || ext == ".hpp" || ext == ".hh" || ext == ".hxx";
}

bool IsFullDirectory(const String& d) {
	if(IsFullPath(d)) {
		FindFile ff(d);
		if(ff) return ff.IsDirectory();
	}
	return false;
}

String PackagePath(const String& name) {
	String uppfile = NativePath(name);
	if(IsFullPath(uppfile)) return NormalizePath(uppfile);
	Vector<String> d = GetUppDirs();
	String p;
	for(int i = 0; i < d.GetCount(); i++) {
		p = NormalizePath(AppendFileName(AppendFileName(d[i], uppfile),
		                  GetFileName(uppfile)) + ".upp");
		if(FileExists(p)) return p;
	}
	return d.GetCount() ? NormalizePath(AppendFileName(AppendFileName(d[0], uppfile),
		                                GetFileName(uppfile)) + ".upp") : "";
}

String SourcePath(const String& package, const String& file) {
	if(IsFullPath(file)) return NativePath(file);
	return NormalizePath(AppendFileName(GetFileFolder(PackagePath(package)), file));
}

String ReadValue(CParser& p) {
	p.Spaces();
	if(p.IsString())
		return p.ReadOneString();
	String v;
	while(p.PeekChar() > ' ' && p.PeekChar() != ',' && p.PeekChar() != ';')
		v.Cat(p.GetChar());
	p.Spaces();
	return v;
}

bool FlagCond::Match(const Vector<String>& conf) const {
	for(int i = 0; i < fc.GetCount(); i++) {
		const FC& f = fc[i];
		if(HasFlag(conf, f.flag) == f.isnot)
			return false;
	}
	return true;
}

void FlagCond::Read(CParser& p) {
	Clear();
	for(;;) {
		bool isnot = p.Char('!');
		if(!p.IsId()) break;
		FC& f = fc.Add();
		f.isnot = isnot;
		f.flag = p.ReadId();
	}
}

String FlagCond::AsString() const {
	String r;
	for(int i = 0; i < fc.GetCount(); i++) {
		const FC& f = fc[i];
		if(i) r << ' ';
		if(f.isnot) r << '!';
		r << f.flag;
	}
	return r;
}

void FlagCond::ReadP(CParser& p) {
	Clear();
	if(!p.Char('(')) return;
	Read(p);
	p.Char(')');
}

String FlagCond::AsStringP() const {
	return fc.GetCount() ? '(' + AsString() + ')' : String();
}

String GetAnyFileName(const char *path)
{
	const char *p = path;
	for(char c; c = *path++;)
		if(c == '\\' || c == '/' || c == ':')
			p = path;
	return p;
}

String GetAnyFileTitle(const char *path)
{
	String fn = GetAnyFileName(path);
	const char *p = fn;
	const char *e = p;
	while(*e && *e != '.')
		e++;
	return String(p, e);
}

String CatAnyPath(String path, const char *more)
{
	if(!more || !*more)
		return path;
	if(!path.IsEmpty() && *path.Last() != '\\' && *path.Last() != '/' &&
	   *more != '\\' && *more != '/')
		path.Cat('/');
	path.Cat(more);
	return path;
}


void SplitHostName(const char *hostname, String& host, int& port) {
	enum { DEFAULT_REMOTE_PORT = 2346 };
	const char *p = hostname;
	while(p && *p != ':')
		p++;
	host = String(hostname, p);
	if(*p++ == ':' && IsDigit(*p))
		port = stou(p);
	else
		port = DEFAULT_REMOTE_PORT;
}

Vector<String> SplitFlags0(const char *flags) {
	return Split(flags, CharFilterTextTest(CharFilterWhitespace));
}

Vector<String> SplitFlags(const char *flags, bool main, const Vector<String>& accepts)
{
	Vector<String> v = SplitFlags0(flags);
	Vector<String> h;
	for(int i = 0; i < v.GetCount(); i++)
		if(v[i][0] == '.') {
			String f = v[i].Mid(1);
			if(main || FindIndex(accepts, f) >= 0)
				h.Add(v[i].Mid(1));
		}
		else
			h.Add(v[i]);
	if(main)
		h.Add("MAIN");
	return h;
}

Vector<String> SplitFlags(const char *flags, bool main)
{
	Vector<String> dummy;
	return SplitFlags(flags, main, dummy);
}

bool HasFlag(const Vector<String>& conf, const char *flag) {
	return FindIndex(conf, flag) >= 0;
}

Vector<String> Combine(const Vector<String>& conf, const char *flags) {
	Vector<String> cfg(conf, 1);
	CParser p(flags);
	while(!p.IsEof()) {
		bool isnot = p.Char('!');
		if(!p.IsId()) break;
		String flg = p.ReadId();
		int i = FindIndex(cfg, flg);
		if(isnot) {
			if(i >= 0) cfg.Remove(i);
		}
		else
			if(i < 0) cfg.Add(flg);
	}
	return cfg;
}

String Gather(const Array<OptItem>& set, const Vector<String>& conf) {
	String s;
	for(int i = 0; i < set.GetCount(); i++)
		if(set[i].Match(conf)) {
			if(!s.IsEmpty()) s.Cat(' ');
			s.Cat(set[i].text);
		}
	return s;
}

int    GetType(const Vector<String>& conf, const char *flags) {
	Vector<String> f = SplitFlags(flags);
	int q = FLAG_UNDEFINED;
	for(int i = 0; i < f.GetCount(); i++)
		if(HasFlag(conf, f[i]))
			if(q == FLAG_UNDEFINED)
				q = i;
			else
				q = FLAG_MISMATCH;
	return q;
}

String RemoveType(Vector<String>& conf, const char *flags)
{
	String old;
	Index<String> f = SplitFlags(flags);
	for(int i = conf.GetCount(); --i >= 0;)
		if(f.Find(conf[i]) >= 0)
		{
			old = conf[i];
			conf.Remove(i);
		}
	return old;
}

int    GetType(const Vector<String>& conf, const char *flags, int def) {
	int q = GetType(conf, flags);
	return q == FLAG_UNDEFINED ? def : q;
}

bool   GetFlag(const Vector<String>& conf, const char *flag) {
	return FindIndex(conf, flag) >= 0;
}

void   OptItem::Load(CParser& p) throw(CParser::Error) {
	ReadP(p);
	text = p.ReadString();
}

String OptItem::AsString() const {
	return AsStringP() + ' ' + AsCString(text);
}

String CustomStep::AsString() const {
	return "custom" + AsStringP() + ' ' + AsCString(ext) + ",\n\t" +
					  AsCString(command, 70, "\t") + ",\n\t" +
					  AsCString(output, 70, "\t") + ";\n\n";
}

void CustomStep::Load(CParser& p) throw(CParser::Error) {
	ReadP(p);
	ext = p.ReadString();
	p.PassChar(',');
	command = p.ReadString();
	p.PassChar(',');
	output = p.ReadString();
	p.PassChar(';');
}

String CustomStep::GetExt() const {
	return ext[0] != '.' ? "." + ToLower(ext) : ToLower(ext);
}

bool   CustomStep::MatchExt(const char *fn) const {
	return ToLower(GetFileExt(fn)) == GetExt();
}

bool when(CParser& p, const char *key, Array<OptItem>& v, String add_fc = Null) {
	if(p.Id(key)) {
		OptItem& itm = v.Add();
		itm.Load(p);
		if(!IsNull(add_fc)) {
			FlagCond::FC& fc = itm.fc.Add();
			fc.isnot = false;
			fc.flag = add_fc;
		}
		p.PassChar(';');
		return true;
	}
	return false;
}

void Package::Load(const char *path) {
	for(;;) {
		charset = 0;
		library.Clear();
		target.Clear();
		flag.Clear();
		option.Clear();
		link.Clear();
		uses.Clear();
		accepts.Clear();
		file.Clear();
		config.Clear();
		custom.Clear();
		unknown.Clear();
		String f = LoadFile(path);
		time = FileGetTime(path);
		CParser p(f);
		try {
			while(!p.IsEof()) {
				if(!when(p, "options", option) &&
				   !when(p, "liboptions", link, "SO") &&
				   !when(p, "link", link) &&
				   !when(p, "library", library) &&
				   !when(p, "flags", flag) &&
				   !when(p, "target", target))
				if(p.Id("charset")) {
					charset = CharsetByName(p.ReadString());
					p.PassChar(';');
				}
				else
				if(p.Id("uses")) {
					do
						uses.Add(ReadValue(p));
					while(p.Char(','));
					p.Char(';');
				}
				else
				if(p.Id("acceptflags")) {
					do
						accepts.Add(ReadValue(p));
					while(p.Char(','));
					p.Char(';');
				}
				else
				if(p.Id("file")) {
					do {
						file.Add(ReadValue(p));
						while(!p.IsChar(',') && !p.IsChar(';')) {
							if(p.Id("readonly"))
								file.Top().readonly = true;
							else
							if(p.Id("separator"))
								file.Top().separator = true;
							else
							if(p.Id("charset"))
								file.Top().charset = CharsetByName(p.ReadString());
							else
							if(p.Id("tabsize"))
								file.Top().tabsize = minmax(p.ReadInt(), 1, 20);
							else
							if(p.Id("font"))
								file.Top().font = minmax(p.ReadInt(), 0, 3);
							else
							if(p.Id("highlight"))
								file.Top().highlight = p.ReadId();
							else
							if(p.Id("options"))
								file.Top().option.Add().Load(p);
							else
							if(p.Id("depends"))
								file.Top().depends.Add(ReadValue(p));
							else {
								const char *q = p.GetPtr();
								if(p.Char('('))
									while(!p.Char(')'))
										p.SkipTerm();
								else
									p.SkipTerm();
								file.Top().unknown.Cat(' ');
								file.Top().unknown.Cat(q, p.GetPtr() - q);
							}
						}
					}
					while(p.Char(','));
					p.PassChar(';');
				}
				else
				if(p.Id("mainconfig")) {
					do {
						String c = p.ReadString();
						p.Char('=');
						p.Id("external");
						p.Id("console");
						p.Id("remotelinux");
						p.Id("normal");
						String f = p.ReadString();
						Config& cf = config.Add();
						cf.name = c;
						cf.param = f;
					}
					while(p.Char(','));
					p.Char(';');
				}
				else
				if(p.Id("custom"))
					custom.Add().Load(p);
				else {
					unknown << p.ReadId() << " ";
					const char *q = p.GetPtr();
					while(!p.Char(';') && !p.IsEof())
						p.SkipTerm();
					unknown.Cat("\r\n\r\n");
					unknown.Cat(q, p.GetPtr());
				}
			}
			return;
		}
		catch(CParser::Error error) {
		prompt:
			switch(Prompt(Ctrl::GetAppName(), CtrlImg::exclamation(),
			              error + "&while parsing package " + DeQtf(path),
				          "Edit \\& Retry", "Ignore",  "Stop")) {
			case 0:
				if(!PromptYesNo("Ignoring will damage package. Everything past the "
					            "point of error will be lost.&Do you want to continue ?"))
					goto prompt;
				Save(path);
				return;
			case 1: {
					TopWindow win;
					LineEdit edit;
					edit.Set(f);
					edit.SetCursor(edit.GetPos(p.GetLine() - 1));
					win.Title(path);
					win.Add(edit.SizePos());
					win.Run();
					SaveFile(path, edit.Get());
				}
				break;
			case -1:
				exit(1);
			}
		}
	}
}

String WriteValue(const String& x) {
	for(const char *s = x; s < x.End(); s++)
		if(byte(*s) <= ' ')
			return AsCString(x);
	return x;
}

void putopt(Stream& out, const char *key, const Array<OptItem>& v) {
	for(int i = 0; i < v.GetCount(); i++)
		out << key << v[i].AsString() << ";\n\n";
}

void putp(Stream& out, const char *key, const Vector<String>& v)
{
	if(v.GetCount()) {
		out << key << "\n";
		for(int i = 0; i < v.GetCount(); i++) {
			if(i) out << ",\n";
			out << '\t' << WriteValue(v[i]);
		}
		out << ";\n\n";
	}
}

bool Package::Save(const char *path) const {
	StringStream out;
	if(charset > 0 && charset < CharsetCount() || charset == CHARSET_UTF8)
		out << "charset " << AsCString(CharsetName(charset)) << ";\n\n";
	putp(out, "uses", uses);
	putopt(out, "flags", flag);
	putp(out, "acceptflags", accepts);
	putopt(out, "target", target);
	putopt(out, "library", library);
	putopt(out, "options", option);
	putopt(out, "link", link);
	if(file.GetCount()) {
		out << "file\n";
		int i;
		for(i = 0; i < file.GetCount(); i++) {
			if(i) out << ",\n";
			const File& f = file[i];
			out << '\t' << WriteValue(f);
			if(f.readonly)
				out << " readonly";
			if(f.separator)
				out << " separator";
			if(f.tabsize > 0)
				out << " tabsize " << f.tabsize;
			if(f.font > 0)
				out << " font " << f.font;
			if(f.charset > 0 && f.charset < CharsetCount() || f.charset == CHARSET_UTF8)
				out << " charset " << AsCString(CharsetName(f.charset));
			if(!IsNull(f.highlight))
				out << " highlight " << f.highlight;
			int i;
			for(i = 0; i < f.depends.GetCount(); i++)
				out << "\n\t\tdepends " << WriteValue(f.depends[i]);
			const Array<OptItem>& m = f.option;
			for(i = 0; i < m.GetCount(); i++)
				out << "\n\t\toptions" << m[i].AsString();
			out << f.unknown;
		}
		out << ";\n\n";
	}
	if(config.GetCount()) {
		out << "mainconfig\n";
		for(int i = 0; i < config.GetCount(); i++) {
			const char *o[] = {
				"normal", "remote", "external", "console"
			};
			const Config& f = config[i];
			if(i) out << ",\n";
			out << '\t' << AsCString(f.name) << " = " << AsCString(f.param);
		}
		out << ";\n\n";
	}
	for(int i = 0; i < custom.GetCount(); i++)
		out << custom[i].AsString();
	out << unknown;
	return SaveChangedFile(path, out.GetResult());
}

/*
void Workspace::Add(const char *prjname) {
	if(package.Find(prjname) >= 0) return;
	Package& prj = package.Add(prjname);
	prj.Load(PackagePath(prjname));
	for(int i = 0; i < prj.uses.GetCount(); i++)
		Add(prj.uses[i]);
}

void Workspace::Scan(const char *prjname) {
	package.Clear();
	Add(prjname);
}
*/
void Workspace::AddLoad(const String& name)
{
	package.Add(name).Load(PackagePath(name));
}

void Workspace::AddUses(Package& p)
{
	int q = package.GetCount();
	for(int i = 0; i < p.uses.GetCount(); i++)
		if(package.Find(p.uses[i]) < 0)
			AddLoad(p.uses[i]);
	for(int i = q; i < package.GetCount(); i++)
		AddUses(package[i]);
}

void Workspace::Scan(const char *prjname) {
	package.Clear();
	AddLoad(prjname);
	AddUses(package[0]);
}

void Workspace::Dump() {
	for(int i = 0; i < package.GetCount(); i++) {
		Package& prj = package[i];
		LOG("Package " << package.GetKey(i));
		LOG(" file ");
		for(int i = 0; i < prj.file.GetCount(); i++)
			LOG("\t" << prj.file[i]);
	}
}

Vector<String> Workspace::GetAllAccepts(int pk) const
{
	Index<String> accepts;
	Index<int> pkg;
	pkg.Add(pk);
	for(int i = 0; i < pkg.GetCount(); i++) {
		const Package& p = package[pkg[i]];
		FindAppend(accepts, p.accepts);
		for(int u = 0; u < p.uses.GetCount(); u++) {
			int f = package.Find(p.uses[u]);
			if(f >= 0)
				pkg.FindAdd(f);
		}
	}
	return accepts.PickKeys();
}
