#include "ide.h"
#pragma hdrstop

#ifdef PLATFORM_POSIX
bool MoveFile(const char *oldpath, const char *newpath)
{
	return !rename(oldpath, newpath);
}
#endif

void OptSetCtrl::Load(const Array<OptItem>& src) {
	Clear();
	for(int i = 0; i < src.GetCount(); i++)
		Add(((FlagCond&)src[i]).AsString(), src[i].text);
}

void OptSetCtrl::Save(Array<OptItem>& dst) {
	dst.Clear();
	for(int i = 0; i < GetCount(); i++) {
		OptItem& m = dst.Add();
		String c = Get(i, 0);
		CParser p(c);
		m.Read(p);
		m.text = Get(i, 1);
	}
}

int FlagFilter(int c) {
	return IsAlNum(c) || c == '_' || c == ' ' ? ToUpper(c) : 0;
}

int FlagFilterM(int c) {
	return c == '.' ? '.' : FlagFilter(c);
}

int CondFilter(int c) {
	return c == '!' ? '!' : FlagFilter(c);
}

OptSetCtrl::OptSetCtrl() {
	Appending().Removing();
	cond.SetFilter(CondFilter);
	AddColumn("When", 10).Edit(cond);
	AddColumn("Value", 15).Edit(text);
}

OptSetCtrl::~OptSetCtrl() {}

void DependsArray::Menu(Bar& bar)
{
	bar.Add("Add package directory file(s)", THISBACK1(AddFile, true));
	bar.Add("Add any file(s)", THISBACK1(AddFile, false));
	bar.Add(file.IsCursor(), "Remove file", THISBACK(RemoveFile));
}

void DependsArray::Sort()
{
	String q = file.GetKey();
	file.Sort();
	file.FindSetCursor(q);
}

void DependsArray::AddFile(bool based)
{
	FileSel& fs = based ? BasedSourceFs() : AnySourceFs();
	BasedSourceFs().BaseDir(PackageDirectory(actualpackage));
	if(!fs.ExecuteOpen("Add file(s) to dependecy list..")) return;
	for(int i = 0; i < fs.GetCount(); i++)
		file.Add(fs[i]);
	file.GoEnd();
	Sort();
	file.WhenArrayAction();
}

void DependsArray::RemoveFile()
{
	file.DoRemove();
}

DependsArray::DependsArray(ArrayCtrl& file)
:	file(file)
{
	file.AddColumn("Aditional dependencies");
	file.WhenBar = THISBACK(Menu);
}

class BaseSetupDlg : public WithBaseSetupLayout<TopWindow>
{
public:
	typedef BaseSetupDlg CLASSNAME;
	BaseSetupDlg();

	bool Run(String& vars);

private:
	void OnUpp();
	void OnBrowseUpp();

private:
	FrameRight<Button> browse_upp;
	SelectDirButton    browse_out;
	SelectDirButton    browse_common;
	bool               new_base;
};

bool BaseSetup(String& vars) { return BaseSetupDlg().Run(vars); }

BaseSetupDlg::BaseSetupDlg()
: browse_out("Output & intermediate files")
, browse_common("Common files")
{
	CtrlLayoutOKCancel(*this, "Assembly setup");
	browse_upp.SetImage(CtrlImg::right_arrow());
	browse_upp <<= THISBACK(OnBrowseUpp);
	upp.AddFrame(browse_upp);
	upp <<= THISBACK(OnUpp);
	browse_out.Attach(output);
	browse_common.Attach(common);
}

bool BaseSetupDlg::Run(String& vars)
{
	upp <<= GetVar("UPP");
	output <<= GetVar("OUTPUT");
	common <<= GetVar("COMMON");
	base <<= vars;
	new_base = IsNull(vars);
	while(TopWindow::Run() == IDOK)
	{
		String varname = ~base;
		String varfile = VarFilePath(varname);
		if(varname != vars)
		{
			if(FileExists(varfile) && !PromptOKCancel(NFormat("Overwrite existing base file [* \1%s\1]?", varfile)))
				continue;
			if(!SaveVars(varname))
			{
				Exclamation(NFormat("Error writing base file [* \1%s\1].", VarFilePath(varname)));
				continue;
			}
		}
		SetVar("UPP", ~upp);
		SetVar("OUTPUT", ~output);
		SetVar("COMMON", ~common);
		Vector<String> paths = SplitDirs(upp.GetText().ToString());
		for(int i = 0; i < paths.GetCount(); i++)
			RealizeDirectory(paths[i]);
		RealizeDirectory(output);
		vars = varname;
		return true;
	}
	return false;
}

void BaseSetupDlg::OnBrowseUpp()
{
	String s = ~upp;
	int b, e;
	if(upp.HasFocus())
		upp.GetSelection(b, e);
	else
		e = s.GetLength();
	b = e;
	while(b > 0 && s[b - 1] != ';')
		b--;
	while(e < s.GetLength() && s[e] != ';')
		e++;
	FileSel fsel;
	String pre = s.Left(b), post = s.Mid(e);
	fsel.ActiveDir(s.Mid(b, e - b));
	if(fsel.ExecuteSelectDir()) {
		String newdir = ~fsel;
		upp <<= pre + newdir + Nvl(post, ";");
		upp.SetWantFocus();
		OnUpp();
	}
}

void BaseSetupDlg::OnUpp()
{
	if(new_base) {
		String s = ~upp;
		int f = s.Find(';');
		if(f >= 0) s.Trim(f);
		base <<= GetFileTitle(s);
	}
}

class SelectPackageDlg : public WithListLayout<TopWindow> {
	typedef SelectPackageDlg CLASSNAME;
public:
	SelectPackageDlg(const char *title, bool selectvars, bool main);

	String         Run(String startwith);

private:
	void           ToolBase(Bar& bar);

	void           OnBaseAdd();
	void           OnBaseEdit();
	void           OnBaseRemove();

	void           OnOK();

	void           OnNew();
	void           OnBase();
	void           OnAll();

	void           Load();
	void           Load(String upp, String dir, int progress_pos, int progres_count, String& case_fixed);
	void           SyncBase(String initvars);
	void           SyncList();

private:
	ArrayCtrl      base;
	FileList       list;

	bool           selectvars;
	bool           loading;
	String         selected;
	int            update;
	Vector<String> packages;
};

SelectPackageDlg::SelectPackageDlg(const char *title, bool selectvars_, bool main)
: selectvars(selectvars_)
{
	CtrlLayoutOKCancel(*this, title);
	Sizeable().Zoomable();
	Icon(IdeImg::Package(), IdeImg::Package());
	base.AutoHideSb();
	base.NoGrid();
	base.AddColumn("Assembly");
	base.WhenCursor = THISBACK(OnBase);
	base.WhenBar = THISBACK(ToolBase);
	base.WhenLeftDouble = THISBACK(OnBaseEdit);
	ok <<= list.WhenLeftDouble = THISBACK(OnOK);
	list.Columns(3);
	splitter.Horz(base, list);
	splitter.SetPos(2000);
	splitter.Zoom(selectvars ? -1 : 1);
	newu <<= THISBACK(OnNew);
	all <<= THISBACK(OnAll);
	all.Show(main);
	all = !main;
	progress.Hide();
}

inline bool PackageLess(String a, String b)
{
	int nc = CompareNoCase(a, b);
	if(nc) return nc < 0;
	return a < b;
};

String SelectPackageDlg::Run(String startwith)
{
	if(!IsSplashOpen())
		Open();
	if(selectvars)
		SyncBase(GetVarsName());
	else
		OnBase();
	ActiveFocus(list);
	list.FindSetCursor(startwith);
	switch(TopWindow::Run())
	{
	case IDOK:  return list.GetCurrentName();
	case IDYES: return selected;
	default:    return Null;
	}
}

void SelectPackageDlg::OnOK()
{
	loading = false;
	AcceptBreak(IDOK);
}

void SelectPackageDlg::OnAll()
{
	OnBase();
}

void SelectPackageDlg::OnBase()
{
	Load();
}

void SelectPackageDlg::OnNew() {
	TemplateDlg dlg;
	LoadFromGlobal(dlg, "NewPackage");
	dlg.Load(GetUppDirs(), !all);
	while(dlg.Run() == IDOK) {
		String nest = ~dlg.nest;
		String name = NativePath(String(~dlg.package));
		String path = AppendFileName(nest, AppendFileName(name, GetFileName(name) + ".upp"));
		if(FileExists(path)) {
			if(PromptOKCancel(NFormat("Use existing package [* \1%s\1]?&"
			                          "(Package template will not be used.)", path)))
			{
				selected = name;
				Break(IDYES);
				break;
			}
			continue;
		}
		RealizePath(path);
		if(!SaveFile(path, Null)) {
			Exclamation(NFormat("Error writing file [* \1%s\1].", path));
			continue;
		}
		dlg.Create();
		selected = name;
		Break(IDYES);
		break;
	}
	StoreToGlobal(dlg, "NewPackage");
}

void SelectPackageDlg::ToolBase(Bar& bar)
{
	bar.Add("New assembly", THISBACK(OnBaseAdd))
		.Key(K_INSERT);
	bar.Add(base.IsCursor(), "Edit assembly", THISBACK(OnBaseEdit))
		.Key(K_CTRL_ENTER);
	bar.Add(base.IsCursor(), "Remove assembly", THISBACK(OnBaseRemove))
		.Key(K_CTRL_DELETE);
}

void SelectPackageDlg::OnBaseAdd()
{
	String vars;
	if(BaseSetup(vars))
		SyncBase(vars);
}

void SelectPackageDlg::OnBaseEdit()
{
	if(!base.IsCursor())
		return;
	String vars = base.Get(0), oldvars = vars;
	if(BaseSetup(vars))
	{
		if(vars != oldvars)
			DeleteFile(VarFilePath(oldvars));
		SyncBase(vars);
	}
}

void SelectPackageDlg::OnBaseRemove()
{
	int c = base.GetCursor();
	if(c < 0)
		return;
	String next;
	if(c + 1 < base.GetCount())
		next = base.Get(c + 1);
	else if(c > 0)
		next = base.Get(c - 1);
	String vars = base.Get(0);
	String varpath = VarFilePath(vars);
	if(PromptOKCancel(NFormat("Remove base file [* \1%s\1]?", varpath)) && !FileDelete(varpath))
		Exclamation(NFormat("Error deleting file [* \1%s\1].", varpath));
	SyncBase(next);
}

void SelectPackageDlg::Load()
{
	update = msecs();
	list.Clear();
	if(selectvars)
	{
		list.Enable(base.IsCursor());
		if(!base.IsCursor())
			return;
		LoadVars((String)base.Get(0));
	}
	Vector<String> upp = GetUppDirs();
	packages.Clear();
	all.Hide();
	progress.Show();
	loading = true;
	String case_fixed;
	for(int i = 0; i < upp.GetCount() && (all || i < 1) && loading; i++) {
		if(!IsSplashOpen() && !IsOpen())
			Open();
		Load(upp[i], Null, 1000 * i, 1000, case_fixed);
	}
	if(!IsNull(case_fixed))
		PromptOK("Case was fixed in some of the files:[* " + case_fixed);
	progress.Hide();
	while(IsSplashOpen())
		ProcessEvents();
	if(!IsOpen())
		Open();
	all.Show();
	if(loading)
	{
		loading = false;
		SyncList();
	}
}

void FixName(const String& dir, const String& name, String& case_fixed)
{
	if(IsFullPath(name))
		return;
	String rp = AppendFileName(dir, name);
	FindFile ff(rp);
	if(!ff || ff.GetName() == GetFileName(name))
		return;
	MoveFile(rp, rp);
	case_fixed << '&' << DeQtf(rp);
}

bool IsMainPackage(const String& dd, const String& nm, String& case_fixed)
{
	Package p;
	p.Load(AppendFileName(dd, nm));
	FixName(dd, nm, case_fixed);
	for(int i = 0; i < p.GetCount(); i++)
		if(!p[i].separator)
			FixName(dd, p[i], case_fixed);
	return p.config.GetCount();
}

void SelectPackageDlg::Load(String upp, String dir, int progress_pos, int progress_count, String& case_fixed)
{
	if(msecs(update) >= 200)
	{
		progress.Set(progress_pos);
		SyncList();
		ProcessEvents();
		if(!loading)
			return;
		update = msecs();
	}

	String d = AppendFileName(upp, dir);
	Vector<String> folders;
	for(FindFile ff(AppendFileName(d, "*.*")); ff; ff.Next())
		if(ff.IsFolder())
			folders.Add(ff.GetName());
	Sort(folders, &PackageLess);
	for(int i = 0; i < folders.GetCount() && loading; i++)
	{
		String fo = folders[i];
		String dd = AppendFileName(d, fo);
		String nm = fo + ".upp";
		String pkg = AppendFileName(dir, fo);
		if(FileExists(AppendFileName(dd, nm)) && (all || IsMainPackage(dd, nm, case_fixed)))
			packages.Add(pkg);
		int ppos = progress_pos + i * progress_count / folders.GetCount();
		int npos = progress_pos + (i + 1) * progress_count / folders.GetCount();
		Load(upp, pkg, ppos, npos - ppos, case_fixed);
	}
}

void SelectPackageDlg::SyncBase(String initvars)
{
	Vector<String> varlist;
	for(FindFile ff(ConfigFile("*.var")); ff; ff.Next())
		if(ff.IsFile())
			varlist.Add(GetFileTitle(ff.GetName()));
	Sort(varlist, &PackageLess);
	base.Clear();
	Append(base, varlist);
	if(!base.FindSetCursor(initvars))
		if(base.GetCount() > 0)
			base.SetCursor(0);
		else
			OnBase();
}

void SelectPackageDlg::SyncList()
{
	Sort(packages, &PackageLess);
	String n = list.GetCurrentName();
	int sc = list.GetSbPos();
	list.Clear();
	for(int i = 0; i < packages.GetCount(); i++)
		list.Add(packages[i], IdeImg::Package());
	list.SetSbPos(sc);
	list.FindSetCursor(n);
}

String SelectPackage(const char *title, const char *startwith, bool selectvars, bool main) {
	return SelectPackageDlg(title, selectvars, main).Run(startwith);
}

const char tempaux[] = "<temp-aux>";
const char prjaux[] = "<prj-aux>";
const char ideaux[] = "<ide-aux>";

Font WorkspaceWork::ListFont()
{
#ifdef PLATFORM_WIN32
	static int q = 0;
	if(!q) {
		FontInfo fi = ScreenInfo().GetFontInfo(StdFont());
		q = fi.GetHeight() - fi.GetInternal();
	}
	return Arial(q);
#else
	return StdFont();
#endif
};

String WorkspaceWork::PackagePathA(const String& pn) {
	if(pn == prjaux) {
		String nm;
		String cfg = ConfigFile("cfg");
		for(const char *s = main; *s; s++)
			nm.Cat(*s == '\\' || *s == '/' ? '$' : *s);
		RealizeDirectory(cfg);
		return AppendFileName(cfg, ForceExt(nm + '@' + GetVarsName(), ".aux"));
	}
	if(pn == ideaux)
		return ConfigFile("ide.aux");
	if(pn == tempaux)
		return ConfigFile(Sprintf("aux%x.tmp",
#ifdef PLATFORM_WIN32
		          GetCurrentProcessId()
#endif
#ifdef PLATFORM_POSIX
		          getpid()
#endif
		       ));
	return PackagePath(pn);
}

void WorkspaceWork::ScanWorkspace() {
	Workspace wspc;
	wspc.Scan(main);
	actualpackage.Clear();
	actualfileindex = -1;
	filelist.Clear();
	package.Clear();
	for(int i = 0; i < wspc.package.GetCount(); i++)
		package.Add(wspc.package.GetKey(i), IdeImg::Package(),
		            i == 0 ? ListFont().Bold() : ListFont(), SBlack, false, 0, Null, SLtBlue);
	if(ide) {
		package.Add(prjaux, IdeImg::PrjAux(), ListFont(), SMagenta);
		package.Add(ideaux, IdeImg::IdeAux(), ListFont(), SMagenta);
		package.Add(tempaux, IdeImg::TempAux(), ListFont(), SMagenta);
	}
	package.SetCursor(0);
	dirtypackage = false;
	SyncWorkspace();
}

void WorkspaceWork::SavePackageData()
{
	String pp = PackagePathA(actualpackage);
	RealizePath(pp);
	actual.Save(pp);
	dirtypackage = false;
}

void WorkspaceWork::SavePackage() {
	if(!actualpackage.IsEmpty() && dirtypackage) {
		actual.charset = (byte)(int)~charset;
		actual.accepts = Split(accepts.GetText().ToString(), ' ');
		option.Save(actual.option);
		link.Save(actual.link);
		library.Save(actual.library);
		flag.Save(actual.flag);
		target.Save(actual.target);
		SavePackageData();
	}
}

void WorkspaceWork::SaveFileOptions()
{
	if(actualfileindex >= 0 && actualfileindex < fileindex.GetCount()) {
		int i = fileindex[actualfileindex];
		if(i >= 0 && i < actual.file.GetCount()) {
			fileoption.Save(actual.file[i].option);
			Vector<String>& dep = actual.file[i].depends;
			dep.Clear();
			for(int i = 0; i < depends.GetCount(); i++)
				dep.Add(depends.Get(i, 0));
		}
	}
}

void WorkspaceWork::FlushPackage() {
	SaveFileOptions();
	actualfileindex = -1;
	SavePackage();
	actualpackage.Clear();
	fileoption.Clear();
	fileoption.Disable();
	option.Clear();
	option.Disable();
	link.Clear();
	link.Disable();
	library.Clear();
	library.Disable();
	flag.Clear();
	flag.Disable();
	target.Clear();
	target.Disable();
	fileoption.Clear();
	fileoption.Disable();
	depends.Clear();
	depends.Disable();
	uses.Clear();
	uses.Disable();
	charset.Disable();
	accepts.Clear();
	accepts.Disable();
	filelist.Clear();
	filelist.Disable();
}

void WorkspaceWork::EnterPackage()
{
	FlushPackage();
	actualpackage = GetActivePackage();
	da.actualpackage = actualpackage;
	if(actualpackage.IsEmpty()) return;
	String pp = PackagePathA(actualpackage);
	RealizePath(pp);
	actual.Load(pp);
	option.Load(actual.option);
	link.Load(actual.link);
	library.Load(actual.library);
	flag.Load(actual.flag);
	target.Load(actual.target);
	int i;
	filelist.Clear();
	fileindex.Clear();
	bool open = true;
	for(i = 0; i < actual.file.GetCount(); i++) {
		Package::File& f = actual.file[i];
		if(f.separator) {
			open = closed.Find(Sepfo(actualpackage, f)) < 0;
			filelist.Add(f, open ? IdeImg::SeparatorClose() : IdeImg::SeparatorOpen(),
			             ListFont().Bold(), open ? SBlue : SBlack, true, 0, Null);
			fileindex.Add(i);
		}
		else
		if(open) {
			filelist.Add(f, IdeFileImage(f), ListFont(), SBlack, false, 0, Null, SLtBlue);
			fileindex.Add(i);
		}
	}
	charset <<= (int)actual.charset;
	uses.Clear();
	for(i = 0; i < actual.uses.GetCount(); i++)
		uses.Add(actual.uses[i], IdeImg::Package(),
		         ListFont(), SBlack, false, 0, Null, SLtBlue);
	String s;
	for(i = 0; i < actual.accepts.GetCount(); i++) {
		if(i) s << ' ';
		s << actual.accepts[i];
	}
	accepts = s.ToWString();
	option.Enable();
	link.Enable();
	library.Enable();
	flag.Enable();
	target.Enable();
	uses.Enable();
	charset.Enable();
	accepts.Enable();
	filelist.Enable();
}

void WorkspaceWork::EnterFile()
{
	SaveFileOptions();
	int i = filelist.GetCursor();
	actualfileindex = -1;
	if(i >= 0 && i < fileindex.GetCount())
		actualfileindex = fileindex[i];
	if(actualfileindex >= 0 && !filelist[i].isdir) {
		fileoption.Load(actual.file[actualfileindex].option);
		fileoption.Enable();
		depends.Enable();
		depends.Clear();
		const Vector<String>& dep = actual.file[actualfileindex].depends;
		for(int i = 0; i < dep.GetCount(); i++)
			depends.Add(dep[i]);
		da.Sort();
	}
	else {
		fileoption.Clear();
		fileoption.Disable();
		depends.Clear();
		depends.Disable();
	}
}

String WorkspaceWork::FileName(int i) const
{
	return i >= 0 && i < fileindex.GetCount() ? (String)actual.file[fileindex[i]] : Null;
}

bool WorkspaceWork::IsSeparator(int i) const
{
	return i >= 0 && i < fileindex.GetCount() ? actual.file[fileindex[i]].separator : true;
}

String WorkspaceWork::GetActiveFileName() const
{
	return FileName(filelist.GetCursor());
}

bool   WorkspaceWork::IsActiveFile() const
{
	int i = filelist.GetCursor();
	return i >= 0 && i < fileindex.GetCount();
}

Package::File& WorkspaceWork::ActiveFile()
{
	return actual.file[fileindex[filelist.GetCursor()]];
}

void WorkspaceWork::SaveLoadPackage() {
	dirtypackage = true;
	SavePackage();
	String p = actualpackage;
	ScanWorkspace();
	package.FindSetCursor(p);
}

void WorkspaceWork::AddFile(ADDFILE af)
{
	String active = GetActivePackage();
	if(active.IsEmpty()) return;
	FileSel fss;
	FileSel *fs = &OutputFs();
	RealizeDirectory(GetCommonDir());
	RealizeDirectory(GetLocalDir());
	switch(af)
	{
	case PACKAGE_FILE: fs = &BasedSourceFs(); fs->BaseDir(GetFileFolder(PackagePathA(active))); break;
	case ANY_FILE:     fs = &AnySourceFs(); break;
	case OUTPUT_FILE:  fs->ActiveDir(GetOutputDir()); break;
#ifdef PLATFORM_POSIX
	case HOME_FILE:    fs->ActiveDir(GetHomeDirectory()); break;
#endif
	case COMMON_FILE:  fs->ActiveDir(GetCommonDir()); break;
	case LOCAL_FILE:   fs->ActiveDir(GetLocalDir()); break;
	}
	if(!fs->ExecuteOpen("Add files to package..")) return;
	int fci = filelist.GetCursor();
	int cs = filelist.GetSbPos();
	int ci = fci >= 0 && fci < fileindex.GetCount() ? fileindex[fci] : -1;
	for(int i = 0; i < fs->GetCount(); i++) {
		Package::File& f = ci >= 0 ? actual.file.Insert(ci++) : actual.file.Add();
		f = (*fs)[i];
		f.readonly = fs->GetReadOnly();
	}
	SaveLoadPackage();
	filelist.SetSbPos(cs);
	filelist.SetCursor(fci >= 0 ? fci : filelist.GetCount() - 1);
	FileSelected();
}

void WorkspaceWork::AddSeparator()
{
	String active = GetActivePackage();
	if(active.IsEmpty()) return;
	String name;
	if(!EditText(name, "Insert separator", "Name"))
		return;
	int fci = filelist.GetCursor();
	int cs = filelist.GetSbPos();
	int ci = fci >= 0 && fci < fileindex.GetCount() ? fileindex[fci] : -1;
	Package::File& f = ci >= 0 ? actual.file.Insert(ci) : actual.file.Add();
	f = name;
	f.separator = f.readonly = true;
	SaveLoadPackage();
	filelist.SetSbPos(cs);
	filelist.SetCursor(fci >= 0 ? fci : filelist.GetCount() - 1);
	FileSelected();
}

void WorkspaceWork::RemoveFile()
{
	int i = filelist.GetCursor();
	int s = filelist.GetSbPos();
	if(i >= 0 && i < fileindex.GetCount()) {
		int fx = fileindex[i];
		if(actual.file[fx].separator && closed.Find(GetActiveSepfo()) >= 0) {
			int px = fx, c;
			while(--px >= 0 && !actual.file[fx].separator)
				;
			if(px >= 0 && (c = closed.Find(Sepfo(GetActivePackage(), actual.file[px]))) >= 0)
				closed.Unlink(c);
		}
		actual.file.Remove(fx);
	}
	SaveLoadPackage();
	filelist.SetSbPos(s);
	filelist.SetCursor(i);
}

void WorkspaceWork::DelFile()
{
	if(!filelist.IsCursor() || filelist[filelist.GetCursor()].isdir) return;
	if(!PromptYesNo("Remove file from package and discard it ?")) return;
	String file = SourcePath(GetActivePackage(), GetActiveFileName());
	RemoveFile();
	::DeleteFile(file);
	dirtypackage = true;
}

void WorkspaceWork::RenameFile()
{
	if(!filelist.IsCursor()) return;
	String n = GetActiveFileName();
	if(!EditText(n, "Rename file", "New name")) return;
	String spath = SourcePath(GetActivePackage(), GetActiveFileName());
	if(!filelist[filelist.GetCursor()].isdir && GetFileLength(spath) >= 0) {
		if(!::MoveFile(spath, SourcePath(GetActivePackage(), n))) {
			Exclamation("Failed to rename the file.&&" + GetErrorMessage(GetLastError()));
			return;
		}
	}
	int i = filelist.GetCursor();
	int s = filelist.GetSbPos();
	if(i >= 0 && i < fileindex.GetCount())
		(String &)actual.file[i] = n;
	SaveLoadPackage();
	filelist.SetSbPos(s);
	filelist.SetCursor(i);
}

void WorkspaceWork::MoveFile(int d)
{
	int fi = filelist.GetCursor();
	int s = filelist.GetSbPos();
	if(fi < 0 || fi >= fileindex.GetCount())
		return;
	int a = fileindex[fi];
	int b = fileindex[fi + d];
	ShowFile(a);
	ShowFile(b);
	if(a < 0 || a >= actual.file.GetCount() || b < 0 || b >= actual.file.GetCount())
		return;
	Swap(actual.file[a], actual.file[b]);
	ShowFile(a);
	ShowFile(b);
	SaveLoadPackage();
	filelist.SetSbPos(s);
	for(int i = 0; i < fileindex.GetCount(); i++)
		if(fileindex[i] == b) {
			filelist.SetCursor(i);
			break;
		}
	filelist.Sync();
}

void WorkspaceWork::MoveFileUp()
{
	MoveFile(-1);
}

void WorkspaceWork::MoveFileDown()
{
	MoveFile(1);
}

WorkspaceWork::Sepfo WorkspaceWork::GetActiveSepfo()
{
	return Sepfo(GetActivePackage(), GetActiveFileName());
}

void WorkspaceWork::GroupOrFile(Point pos)
{
	if(pos.x < filelist.GetIconWidth())
		Group();
	if(filelist.IsCursor() && !filelist[filelist.GetCursor()].isdir)
		FileSelected();
}

void   WorkspaceWork::Group()
{
	if(!filelist.IsCursor() || !filelist[filelist.GetCursor()].isdir)
		return;
	int i = filelist.GetCursor();
	int s = filelist.GetSbPos();
	Sepfo as = GetActiveSepfo();
	if(closed.Find(as) >= 0)
		closed.UnlinkKey(as);
	else
		closed.Put(as);
	SaveLoadPackage();
	filelist.SetSbPos(s);
	filelist.SetCursor(i);
}

void WorkspaceWork::OpenAllGroups()
{
	for(int i = 0; i < actual.file.GetCount(); i++)
		if(actual.file[i].separator)
			closed.UnlinkKey(Sepfo(GetActivePackage(), actual.file[i]));
	SaveLoadPackage();
}

void WorkspaceWork::CloseAllGroups()
{
	for(int i = 0; i < actual.file.GetCount(); i++)
		if(actual.file[i].separator)
			closed.FindPut(Sepfo(GetActivePackage(), actual.file[i]));
	SaveLoadPackage();
}

void  WorkspaceWork::ShowFile(int pi)
{
	bool open = true;
	Sepfo sf;
	for(int i = 0; i < actual.file.GetCount(); i++) {
		if(actual.file[i].separator) {
			sf = Sepfo(GetActivePackage(), actual.file[i]);
			open = closed.Find(sf) < 0;
		}
		else
		if(i == pi) {
			if(!open) {
				int i = filelist.GetCursor();
				int s = filelist.GetSbPos();
				closed.UnlinkKey(sf);
				SaveLoadPackage();
				filelist.SetSbPos(s);
				filelist.SetCursor(i);
			}
			return;
		}
	}
}


void WorkspaceWork::FileMenu(Bar& menu)
{
	bool sel = filelist.IsCursor() && filelist[filelist.GetCursor()].isdir;
	bool isaux = (actualpackage == tempaux || actualpackage == prjaux || actualpackage == ideaux);
	menu.Add(!isaux, "Insert package directory file(s)", THISBACK1(AddFile, PACKAGE_FILE))
		.Help("Insert file relative to current package");
	menu.Add("Insert separator", THISBACK(AddSeparator))
		.Help("Insert text separator line");
	menu.Separator();
	menu.Add("Insert any file(s)", THISBACK1(AddFile, ANY_FILE))
		.Key(K_CTRL_I)
		.Help("Insert files from anywhere on disk (discouraged in portable packages)");
	menu.Add(isaux, "Insert output directory file(s)", THISBACK1(AddFile, OUTPUT_FILE))
		.Help("Open file selector in output / intermediate directory for current package");
	menu.Add(isaux, "Insert Common directory file(s)", THISBACK1(AddFile, COMMON_FILE))
		.Help("Open file selector in Common directory for current package");
	menu.Add(isaux, "Insert Local directory file(s)", THISBACK1(AddFile, LOCAL_FILE))
		.Help("Open file selector in Local directory for current package");
#ifdef PLATFORM_POSIX
	menu.Add("Insert home directory file(s)", THISBACK1(AddFile, HOME_FILE))
		.Help("Open file selector in current user's HOME directory");
#endif
	menu.Separator();
	if(sel)
		menu.Add(closed.Find(GetActiveSepfo()) < 0 ? "Close group\t[double-click]"
		                                           : "Open group\t[double-click]", THISBACK(Group));
	menu.Add("Open all groups", THISBACK(OpenAllGroups));
	menu.Add("Close all groups", THISBACK(CloseAllGroups));
	menu.Separator();
	BuildFileMenu(menu);
	menu.Separator();
	menu.Add(filelist.IsCursor(), sel ? "Rename separator" : "Rename file", THISBACK(RenameFile))
		.Help("Change separator text");
	menu.Add(filelist.IsCursor(), sel ? "Remove separator " : "Remove file", THISBACK(RemoveFile))
		.Key(ide ? K_ALT_DELETE : K_DELETE)
		.Help("Remove file / separator from package");
	menu.Add(filelist.IsCursor() && !sel, "Delete file", THISBACK(DelFile))
		.Help("Remove file reference from package & delete file on disk");
	menu.Separator();
	menu.Add(filelist.GetCursor() > 0, "Move up", THISBACK(MoveFileUp))
		.Key(ide ? K_SHIFT_CTRL_UP : K_CTRL_UP)
		.Help("Move current file one position towards package beginning");
	menu.Add(filelist.IsCursor() && filelist.GetCursor() < filelist.GetCount() - 1, "Move down",
	         THISBACK(MoveFileDown))
		.Key(ide ? K_SHIFT_CTRL_DOWN : K_CTRL_DOWN)
		.Help("Move current file one position towards package end");
	FilePropertiesMenu(menu);
}

void WorkspaceWork::UpdateUses()
{
	int i = uses.GetCursor();
	SaveLoadPackage();
	if(i >= 0)
		uses.SetCursor(i);
}

void WorkspaceWork::AddNormalUses()
{
	String p = SelectPackage("Select package");
	if(p.IsEmpty()) return;
	actual.uses.Add(p);
	UpdateUses();
}

void WorkspaceWork::RemoveUses()
{
	int i = uses.GetCursor();
	if(i >= 0)
		actual.uses.Remove(i);
	UpdateUses();
}

void WorkspaceWork::Move(Vector<String>& v, FileList& ta, int d)
{
	int i = ta.GetCursor();
	if(i < 0 || i + d < 0 || i + d > v.GetCount() - 1) return;
	Swap(v[i], v[i + d]);
	SaveLoadPackage();
	ta.SetCursor(i + d);
}

void WorkspaceWork::MoveUsesUp()
{
	Move(actual.uses, uses, -1);
}

void WorkspaceWork::MoveUsesDown()
{
	Move(actual.uses, uses, 1);
}

void WorkspaceWork::Goto()
{
	if(uses.IsCursor())
		package.FindSetCursor(uses.GetCurrentName());
}

void WorkspaceWork::UsesMenu(Bar& menu)
{
	menu.Add("Add package", IdeImg::package_add(), THISBACK(AddNormalUses));
	menu.Add(uses.IsCursor(), "Remove uses", THISBACK(RemoveUses)).Key(K_DELETE);
	menu.Separator();
	menu.Add(uses.GetCursor() > 0, "Move up", THISBACK(MoveUsesUp)).Key(K_CTRL_UP);
	menu.Add(uses.IsCursor() && uses.GetCursor() < uses.GetCount() - 1, "Move down",
		     THISBACK(MoveUsesDown)).Key(K_CTRL_DOWN);
	menu.Separator();
	menu.Add(uses.IsCursor(), "Go to package", THISBACK(Goto));
}

void WorkspaceWork::RemovePackageMenu(Bar& bar)
{
	if(bar.IsScanKeys() || bar.IsScanHelp() || !bar.IsMenuBar())
		return;
	String active = GetActivePackage();
	int usecnt = 0;
	for(int i = 0; i < package.GetCount(); i++)
	{
		Package prj;
		String pp = PackagePath(package[i].name);
		prj.Load(pp);
		if(FindIndex(prj.uses, active) >= 0)
		{
			usecnt++;
			bar.Add("Remove from '" + UnixPath(package[i].name) + '\'', THISBACK1(RemovePackage, package[i].name))
				.Help(NFormat("Remove package '%s' from uses section in '%s'", active, pp));
		}
	}
	if(usecnt > 1)
	{
		bar.MenuSeparator();
		bar.Add("Remove all uses", THISBACK1(RemovePackage, String(Null)))
			.Help(NFormat("Remove package '%s' from all uses in active project and its submodules", active));
	}
}

void WorkspaceWork::RemovePackage(String from_package)
{
	String active = GetActivePackage();
	if(IsNull(from_package) && !PromptYesNo(NFormat(
		"Remove package [* \1%s\1] from uses sections of all current packages ?", active)))
		return;
	for(int i = 0; i < package.GetCount(); i++)
		if(IsNull(from_package) || package[i].name == from_package)
		{
			String pp = PackagePath(package[i].name);
			RealizePath(pp);
			Package prj;
			prj.Load(pp);
			bool dirty = false;
			for(int i = prj.uses.GetCount(); --i >= 0;)
				if(prj.uses[i] == active)
				{
					prj.uses.Remove(i);
					dirty = true;
				}
			if(dirty) prj.Save(pp);
		}
	ScanWorkspace();
}

void WorkspaceWork::PackageMenu(Bar& menu)
{
	if(!menu.IsScanKeys()) {
		String act = UnixPath(GetActivePackage());
		menu.Add(~NFormat("Add package to '%s'", act), IdeImg::package_add(), THISBACK(AddNormalUses));
		RemovePackageMenu(menu);
	}
}

WorkspaceWork::WorkspaceWork()
:	da(depends)
{
	package.WhenEnterItem = package.WhenKillCursor = THISBACK(EnterPackage);
	package.WhenBar = THISBACK(PackageMenu);
	package.NoRoundSize();
	filelist.WhenEnterItem = package.WhenKillCursor = THISBACK(EnterFile);
	filelist.WhenLeftClickPos = THISBACK(GroupOrFile);
	filelist.WhenLeftDouble = THISBACK(Group);
	filelist.WhenBar = THISBACK(FileMenu);
	filelist.Columns(3);
	filelist.NoRoundSize();
	DlCharsetD(charset);
	uses.WhenBar = THISBACK(UsesMenu);
	uses.WhenLeftDouble = THISBACK(Goto);
	uses.NoRoundSize();
	package_splitter.Vert(package, uses);
	pk_opt_splitter.Horz(package_splitter, accepts_charset);
	pk_opt_splitter.SetPos(2000);
	pk_opt_splitter.SetMin(1, 4000);
	fileopt_splitter.Vert(fileoption, depends);
	file_splitter.Horz(filelist, fileopt_splitter);
	file_splitter.SetPos(4000);
	vert_splitter.Vert(pk_opt_splitter, file_splitter);
	vert_splitter.SetPos(7000);
	opt_splitter.Vert();
	opt_splitter << flag << target << library << option << link;
	option.Name("Compiler options");
	link.Name("Link options");
	fileoption.Name("Compiler options");
	library.Name("Aditional libraries");
	flag.Name("Add or remove flag(s)");
	target.Name("Target file override");
	option.Disable();
	library.Disable();
	fileoption.Disable();
	link.Disable();
	uses.Disable();
	charset.Disable();
	filelist.Disable();
	actualfileindex = -1;
	dirtypackage = false;
	accepts.SetFilter(FlagFilter);
	accepts <<=
	charset <<=
	flag.WhenArrayAction =
	target.WhenArrayAction =
	library.WhenArrayAction =
	link.WhenArrayAction =
	option.WhenArrayAction =
	fileoption.WhenArrayAction =
	depends.WhenArrayAction = THISBACK(SetDirty);
	accepts_charset << opt_splitter.VSizePos(30, 0);
}

void WorkspaceWork::TableLabels()
{
	package.AddFrame(package_label);
	package_label.SetLabel("Package");
	package_label.NoTransparent();
	package_label.Height(package_label.GetStdSize().cy + 4);
	package_label.SetFrame(ButtonFrame());
	uses.AddFrame(uses_label);
	uses_label.SetLabel("Uses");
	uses_label.NoTransparent();
	uses_label.Height(uses_label.GetStdSize().cy + 4);
	uses_label.SetFrame(ButtonFrame());
	filelist.AddFrame(file_label);
	file_label.SetLabel("File");
	file_label.NoTransparent();
	file_label.Height(file_label.GetStdSize().cy + 4);
	file_label.SetFrame(ButtonFrame());
}

void WorkspaceWork::SerializeClosed(Stream& s)
{
	Workspace wspc;
	wspc.Scan(main);
	Vector<Sepfo> list;
	for(int i = 0; i < wspc.GetCount(); i++) {
		String pk = wspc[i];
		const Package& p = wspc.GetPackage(i);
		for(int i = 0; i < p.GetCount(); i++)
			if(p[i].separator) {
				Sepfo sf(pk, p[i]);
				if(closed.Find(sf) >= 0)
					list.Add(sf);
			}
	}
	s % list;
	closed = list;
}

struct PackageEditor : WorkspaceWork, TopWindow {
	virtual bool Key(dword key, int);

	WithUppLayout<StaticRect> acc;

	void Serialize(Stream& s);

	PackageEditor();
};

bool PackageEditor::Key(dword key, int count)
{
	if(key == K_ESCAPE) {
		Close();
		return true;
	}
	return TopWindow::Key(key, count);
}

PackageEditor::PackageEditor() {
	Title("Package organizer");
	ide = false;
	InitLayout(accepts_charset, acc, *this, acc);
	acc.SizePos();
	Add(vert_splitter);
	SetRect(0, 0, 750, 500);
	SetMinSize(Size(750, 500));
}

void PackageEditor::Serialize(Stream& s) {
	int version = 1;
	s / version;
	s % filelist % uses % package;
	SerializePlacement(s);
	if(version >= 1)
		s % vert_splitter % pk_opt_splitter % package_splitter
		% opt_splitter % file_splitter % fileopt_splitter;
}

void EditPackages(const char *main, const char *startwith, String& cfg) {
	PackageEditor pe;
	pe.TableLabels();
	pe.Sizeable();
	StringStream in(cfg);
	pe.Serialize(in);
	pe.SetMain(main);
	pe.ScanWorkspace();
	pe.FindSetPackage(startwith);
	pe.Run();
	pe.FlushPackage();
	StringStream out;
	pe.Serialize(out);
	cfg = out;
}
