struct FileTemplate {
	String filename;
	String condition;
	String text;
};

enum { TEMPLATEITEM_ID, TEMPLATEITEM_FILENAME, TEMPLATEITEM_OPTION, TEMPLATEITEM_SELECT };

struct TemplateItem {
	String         label;
	int            type;
	String         id;
	Vector<String> value;
	String         init;
};

struct PackageTemplate {
	String              name;
	bool                main, sub;
	Array<TemplateItem> item;
	Array<FileTemplate> file;
};

struct AppPreview : Ctrl {
	virtual void Paint(Draw& w);
	virtual void Layout();
	virtual void MouseWheel(Point p, int zdelta, dword keyflags);

	ScrollBar      sb;
	struct Line : Moveable<Line> {
		String text;
		bool   header;
	};
	Vector<Line> line;

	void Scroll() { Refresh(); }

	void Clear() { line.Clear(); Refresh(); sb.Set(0); }
	void Add(const String& data, bool header = false);

	typedef AppPreview CLASSNAME;

	AppPreview();
	virtual ~AppPreview();
};

struct TemplateDlg : public WithNewPackageLayout<TopWindow> {
	AppPreview      preview;

	Array<PackageTemplate> pt;

	Array<Label>  label;
	Array<Ctrl>   ctrl;
	DelayCallback delay;

	const PackageTemplate& ActualTemplate();

	String GetPackageFile();
	void Preview();
	void EnterTemplate();
	void LoadNest(const char *dir, bool main);
	void Load(const Vector<String>& path, bool main);
	void EnableCreate() { ok.Enable(!IsNull(package)); Inits(); delay.Invoke(); }
	void Inits();

	ArrayMap<String, EscValue> MakeVars0();
	ArrayMap<String, EscValue> MakeVars();
	void Create();

	void Serialize(Stream& s) { SerializePlacement(s); }

	typedef TemplateDlg CLASSNAME;
	TemplateDlg();
	virtual ~TemplateDlg();
};

String SelectPackage(const char *title, const char *startwith = NULL,
	bool selectvars = false, bool all = false);

int CondFilter(int c);
int FlagFilter(int c);
int FlagFilterM(int c);

class OptSetCtrl : public ArrayCtrl {
	EditField  cond;
	EditString text;

public:
	void Load(const Array<OptItem>& src);
	void Save(Array<OptItem>& dst);

	void Name(const char *name)            { HeaderTab(1).SetText(name); }

	typedef OptSetCtrl CLASSNAME;

	OptSetCtrl();
	virtual ~OptSetCtrl();
};

class DependsArray {
	ArrayCtrl& file;

	void AddFile(bool based);
	void RemoveFile();
	void Menu(Bar& menu);

public:
	typedef DependsArray CLASSNAME;

	String     actualpackage;

	void Sort();

	DependsArray(ArrayCtrl& file);
};

struct WorkspaceWork {
	static    Font ListFont();

	bool        dirtypackage;
	FileList    package;
	FileList    filelist;
	Vector<int> fileindex;
	FileList    uses;

	EditString accepts;
	DropList   charset;
	FrameTop<Label> package_label;
	FrameTop<Label> uses_label;
	FrameTop<Label> file_label;
	StaticRect accepts_charset;
	Splitter   opt_splitter;
	Splitter   fileopt_splitter;
	Splitter   file_splitter;
	Splitter   vert_splitter;
	Splitter   pk_opt_splitter;
	Splitter   package_splitter;
	OptSetCtrl flag;
	OptSetCtrl target;
	OptSetCtrl library;
	OptSetCtrl link;
	OptSetCtrl option;
	OptSetCtrl fileoption;

	ArrayCtrl    depends;
	DependsArray da;

	String    main;
	String    actualpackage;
	int       actualfileindex;
	Package   actual;

	struct Sepfo : Moveable<Sepfo> {
		String package;
		String separator;

		bool operator==(const Sepfo& s) const { return package == s.package && separator == s.separator; }
		unsigned GetHashValue() const         { return CombineHash(::GetHashValue(package),
		                                                           ::GetHashValue(separator)); }
		void Serialize(Stream& s)             { s % package % separator; }
		Sepfo(const String& package, const String& separator) : package(package), separator(separator) {}
		Sepfo() {}
	};

	Index<Sepfo> closed;

	bool        ide;

	Sepfo  GetActiveSepfo();
	void   Group();
	void   OpenAllGroups();
	void   CloseAllGroups();
	void   GroupOrFile(Point pos);

	virtual void   EnterPackage();
	virtual void   EnterFile();
	virtual void   BuildFileMenu(Bar& bar)            {}
	virtual void   FilePropertiesMenu(Bar& bar)       {}
	virtual String GetOutputDir()                     { return Null; }
	virtual void   SyncWorkspace()                    {}
	virtual void   FileSelected()                     {}

	void   SetIde()                                   { ide = true; }
	void   SerializeFileSetup(Stream& s)              { s % filelist % uses % package; }

	void   SetMain(const String& m)                   { main = m; }
	void   FindSetPackage(const String& s)            { package.FindSetCursor(s); }

	void   ShowFile(int pi);

	void   SetDirty()                                 { dirtypackage = true; }
	void   FlushPackage();
	void   SaveFileOptions();

	String         GetActivePackage() const                   { return package.GetCurrentName(); }
	String         GetActiveFileName() const;
	bool           IsActiveFile() const;
	Package::File& ActiveFile();
	String         FileName(int i) const;
	bool           IsSeparator(int i) const;

	void ScanWorkspace();

	void SavePackage();
	void SavePackageData();

	void SaveLoadPackage();

	void MoveFile(int d);
	void Move(Vector<String>& v, FileList& ta, int d);

	enum ADDFILE { PACKAGE_FILE, OUTPUT_FILE, HOME_FILE, COMMON_FILE, LOCAL_FILE, ANY_FILE };
	void AddFile(ADDFILE type);
	void AddSeparator();
	void RemoveFile();
	void MoveFileUp();
	void MoveFileDown();
	void DelFile();
	void RenameFile();

	void UpdateUses();
	void AddNormalUses();
	void AddAnyUses();
	void RemoveUses();
	void MoveUsesUp();
	void MoveUsesDown();
	void Goto();

	void RemovePackageMenu(Bar& bar);
	void RemovePackage(String from_package);

	void PackageMenu(Bar& bar);
	void UsesMenu(Bar& bar);
	void FileMenu(Bar& bar);

	String PackagePathA(const String& pn);

	void SerializeClosed(Stream& s);

	void TableLabels();

	typedef WorkspaceWork CLASSNAME;

	WorkspaceWork();
};

void EditPackages(const char *main, const char *startwith, String& cfg);
